"""Test the numeric functions."""

import attoworld as aw
import numpy as np


def test_fornberg_stencil():
    """Test that the fornberg stencil is correct."""
    np.testing.assert_allclose(
        aw.numeric.fornberg_stencil(2, np.array([-1.0, 0.0, 1.0])),
        np.array([1.0, -2.0, 1.0]),
    )


def test_fwhm():
    """Check that the fwhm of a Gaussian is as expected."""
    x = np.linspace(-5, 5, 65)
    dx = x[1] - x[0]
    y = np.exp(-(x**2) / 2)
    fwhm = aw.numeric.fwhm(y, dx)
    np.testing.assert_allclose(fwhm, 2.3547421160236)



"""Classes for organizing data from a Frequency Resolved Optical Gating measurement."""

from dataclasses import dataclass
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from scipy import constants

from ..plot import label_letter
from .decorators import yaml_io
from .interop import ComplexSpectrum, Waveform
from .spectrogram import Spectrogram


@yaml_io
@dataclass(slots=True)
class FrogData:
    """Stores data from a FROG measurement.

    Attributes:
        spectrum (ComplexSpectrum): the reconstructed complex spectrum
        pulse (Waveform): time-domain reconstructed field
        measured_spectrogram (Spectrogram): measured (binned) data
        reconstructed_spectrogram (Spectrogram): spectrogram resulting from reconstructed field
        raw_reconstruction (np.ndarray): the raw reconstructed time-domain pulse of the FROG
        f0: the central frequency of the spectrum
        dt: the time step

    """

    spectrum: ComplexSpectrum
    pulse: Waveform
    measured_spectrogram: Spectrogram
    reconstructed_spectrogram: Spectrogram
    raw_reconstruction: np.ndarray
    f0: float
    dt: float

    def lock(self):
        """Make the data immutable."""
        self.raw_reconstruction.setflags(write=False)
        self.spectrum.lock()
        self.pulse.lock()
        self.measured_spectrogram.lock()
        self.reconstructed_spectrogram.lock()

    def save(self, base_filename):
        """Save in the Trebino FROG format.

        Args:
            base_filename: base of the file path; 4 files will be made from it: .A.dat, .Arecon.dat, .Ek.dat, and .Speck.dat

        """
        self.measured_spectrogram.save(base_filename + ".A.dat")
        self.reconstructed_spectrogram.save(base_filename + ".Arecon.dat")

        t = 1e15 * self.dt * np.array(range(len(self.raw_reconstruction)))
        t -= np.mean(t)
        f = np.fft.fftshift(np.fft.fftfreq(len(t), d=self.dt)) + self.f0
        lam = constants.speed_of_light / f
        raw_spec = np.fft.fftshift(np.fft.fft(self.raw_reconstruction))

        with open(base_filename + ".Ek.dat", "w") as time_file:
            for _i in range(len(self.raw_reconstruction)):
                time_file.write(
                    f"{t[_i]:.15g}\t{np.abs(self.raw_reconstruction[_i]) ** 2:.15g}\t{np.angle(self.raw_reconstruction[_i]):.15g}\t{np.real(self.raw_reconstruction[_i]):.15g}\t{np.imag(self.raw_reconstruction[_i]):.15g}\n"
                )

        with open(base_filename + ".Speck.dat", "w") as spec_file:
            for _i in range(len(self.raw_reconstruction)):
                spec_file.write(
                    f"{lam[_i]:.15g}\t{np.abs(raw_spec[_i]) ** 2:.15g}\t{np.angle(raw_spec[_i]):.15g}\t{np.real(raw_spec[_i]):.15g}\t{np.imag(raw_spec[_i]):.15g}\n"
                )

    def plot_measured_spectrogram(self, ax: Optional[Axes] = None, log: bool = False):
        """Plot the measured spectrogram.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            log (bool): plot on log scale

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        if log:
            self.measured_spectrogram.plot_log(ax)
        else:
            self.measured_spectrogram.plot(ax)
        ax.set_title("Measured")
        return fig

    def plot_reconstructed_spectrogram(
        self, ax: Optional[Axes] = None, log: bool = False
    ):
        """Plot the reconstructed spectrogram.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            log (bool): plot on log scale

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        if log:
            self.reconstructed_spectrogram.plot_log(ax)
        else:
            self.reconstructed_spectrogram.plot(ax)
        ax.set_title(
            f"Retrieved (G': {self.get_error():0.2e}; G: {self.get_G_error():0.2e})"
        )
        return fig

    def plot_pulse(
        self, ax: Optional[Axes] = None, phase_blanking: float = 0.05, xlim=None
    ):
        """Plot the reconstructed pulse.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            phase_blanking: only show phase information (instantaneous frequency) above this level relative to max intensity
            xlim: pass arguments to set_xlim() to constrain the x-axis

        """
        return self.pulse.to_complex_envelope().plot(ax, phase_blanking, xlim)

    def plot_spectrum(
        self, ax: Optional[Axes] = None, phase_blanking: float = 0.05, xlim=None
    ):
        """Plot the reconstructed spectrum and group delay curve.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            phase_blanking: only show phase information (group delay) above this level relative to max intensity
            xlim: pass arguments to set_xlim() to constrain the x-axis

        """
        return self.spectrum.to_intensity_spectrum().plot_with_group_delay(
            ax, phase_blanking=phase_blanking, shift_from_centered=True, xlim=xlim
        )

    def plot_all(
        self,
        phase_blanking=0.05,
        time_xlims=None,
        wavelength_autoscale=1e-3,
        wavelength_xlims=None,
        figsize=None,
        log: bool = False,
    ):
        """Produce a 4-panel plot of the FROG results, combining calls to plot_measured_spectrogram(),
        plot_reconstructed_spectrogram(), plot_pulse() and plot_spectrum() as subplots, with letter labels.

        Args:
            phase_blanking: relative intensity at which to show phase information
            time_xlims: x-axis limits to pass to the plot of the pulse
            wavelength_autoscale: intensity relative to the peak to include within the spectrum (overrides wavelength_xlims unless set to None)
            wavelength_xlims: x-axis limits to pass to the plot of the spectrum
            figsize: custom figure size
            log (bool): plot on log scale

        """
        if figsize is None:
            default_figsize = plt.rcParams["figure.figsize"]
            figsize = (default_figsize[0] * 2, default_figsize[1] * 2)
        fig, ax = plt.subplots(2, 2, figsize=figsize)
        self.plot_measured_spectrogram(ax[0, 0], log=log)
        label_letter("a", ax[0, 0])
        self.plot_reconstructed_spectrogram(ax[1, 0], log=log)
        label_letter("b", ax[1, 0])
        self.plot_pulse(ax[0, 1], xlim=time_xlims, phase_blanking=phase_blanking)
        label_letter("c", ax[0, 1])
        if wavelength_autoscale is not None:
            if not isinstance(wavelength_xlims, tuple):
                spec = self.spectrum.to_intensity_spectrum()
                wl_nm = spec.wavelength_nm()
                indices = np.where(
                    spec.spectrum / np.max(spec.spectrum) > wavelength_autoscale
                )[0]
                wavelength_xlims = (wl_nm[indices[-1]], wl_nm[indices[0]])
        self.plot_spectrum(
            ax[1, 1], xlim=wavelength_xlims, phase_blanking=phase_blanking
        )
        label_letter("d", ax[1, 1])
        return fig

    def get_error(self) -> float:
        """Get the G' error of the reconstruction."""
        norm_measured = np.linalg.norm(self.measured_spectrogram.data)
        norm_retrieved = np.linalg.norm(self.reconstructed_spectrogram.data)
        return np.sqrt(
            np.sum(
                (
                    self.measured_spectrogram.data[:] / norm_measured
                    - self.reconstructed_spectrogram.data[:] / norm_retrieved
                )
                ** 2
            )
            / np.sum((self.measured_spectrogram.data[:] / norm_measured) ** 2)
        )

    def get_G_error(self) -> float:
        """Get the G (note: no apostrophe) error. This one doesn't mean much, but is useful
        for comparing reconstructions of the same spectrogram between different programs.
        """
        return np.sqrt(
            (1.0 / float(len(self.measured_spectrogram.data[:]) ** 2))
            * np.sum(
                (
                    self.measured_spectrogram.data[:]
                    / np.max(self.measured_spectrogram.data[:])
                    - self.reconstructed_spectrogram.data[:]
                    / np.max(self.reconstructed_spectrogram.data[:])
                )
                ** 2
            )
        )

    def get_fwhm(self) -> float:
        """Get the full-width-at-half-max value of the reconstructed pulse."""
        return self.pulse.get_envelope_fwhm()



"""This module will contain personal modules from our coworkers, allowing for easier sharing
of data processing files that utilize functions that might not otherwise be of general use.
"""

from . import marco, nick, vlad

__all__ = ["nick", "vlad", "marco"]



"""Contain data describing a spectrogram."""

from dataclasses import dataclass
from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from scipy import constants

from ..numeric import (
    block_binning_1d,
    block_binning_2d,
    interpolate,
)
from .decorators import yaml_io


@yaml_io
@dataclass(slots=True)
class FrogBinSettings:
    """Stores the settings for binning a FROG measurement."""

    size: int
    dt: float
    f0: float
    t0: float
    auto_t0: bool
    dc_offset: float
    time_binning: int
    freq_binning: int
    median_binning: bool
    spatial_chirp_correction: bool


@yaml_io
@dataclass(slots=True)
class Spectrogram:
    """Contains the data describing a spectrogram.

    Attributes:
        data (np.ndarray): 2d spectrogram
        time (np.ndarray): time vector
        freq (np.ndarray): frequency vector

    """

    data: np.ndarray
    time: np.ndarray
    freq: np.ndarray

    def lock(self):
        """Make the data immutable."""
        self.data.setflags(write=False)
        self.time.setflags(write=False)
        self.freq.setflags(write=False)

    def save(self, filename):
        """Save in the .A.dat file format used by FROG .etc
        Args:
            filename: the file to be saved.

        The file is structured like this:
        [number of wavelengths] [number of times]
        [minimum value of the trace] [maximum value of the trace]
        [array of wavelengths]
        [array of times]
        [data array as single column]
        """
        with open(filename, "w") as file:
            file.write(f"{len(self.freq)}\t{len(self.time)}\n")
            file.write(f"{np.min(self.data[:])}\t{np.max(self.data[:])}\n")
            for freq in self.freq:
                wavelength_nm = 1e9 * constants.speed_of_light / freq
                file.write(f"{wavelength_nm:15.15g}\n")
            for time in self.time:
                time_fs = 1e15 * time
                file.write(f"{time_fs:15.15g}\n")
            for x in self.data:
                for y in x:
                    file.write(f"{y:15.15g}\n")

    def to_block_binned(self, freq_bin: int, time_bin: int, method: str = "mean"):
        """Apply block-binning to the spectrogram.

        Args:
            freq_bin (int): block size for averaging in the frequency direction
            time_bin (int): block size for averaging in the time-direction
            method (str): can be ```mean``` or ```median```

        """
        return Spectrogram(
            data=block_binning_2d(self.data, time_bin, freq_bin, method),
            freq=block_binning_1d(self.freq, freq_bin, "mean"),
            time=block_binning_1d(self.time, time_bin, "mean"),
        )

    def to_per_frequency_dc_removed(self, extra_offset: float = 0.0):
        """Perform DC offset removal on a measured spectrogram, on a per-frequency basis.

        Args:
            extra_offset (float): subtract a value from the entire array (negative values are always set to zero)

        Returns:
            Spectrogram: the spectrogram with offset removed.

        """
        new_data = np.array(self.data)
        new_data -= extra_offset
        new_data[new_data < 0.0] = 0.0
        for _i in range(new_data.shape[0]):
            new_data[_i, :] -= np.min(new_data[_i, :])

        return Spectrogram(data=new_data, time=self.time, freq=self.freq)

    def to_symmetrized(self):
        """Average the trace with a time-reversed copy. This might be useful for getting a reconstruction of difficult data, but keep in mind that the resulting measured trace will no longer represent the real measurement and should not be published as such."""
        return Spectrogram(
            data=0.5 * (self.data + np.fliplr(self.data)),
            time=self.time,
            freq=self.freq,
        )

    def to_removed_spatial_chirp(self):
        """Remove the effects of spatial chirp on an SHG-FROG trace by centering all single-frequency autocorrelations to the same time-zero."""
        new_data = np.array(self.data)
        for i in range(len(self.freq)):
            total = np.sum(self.data[i, :])
            if total > 0.0:
                t0 = np.sum(self.time * self.data[i, :]) / total
                new_data[i, :] = interpolate(self.time + t0, self.time, self.data[i, :])

        return Spectrogram(data=new_data, time=self.time, freq=self.freq)

    def to_combined_and_binned(
        self,
        other,
        stitching_band: Tuple[float, float],
        dim: int = 64,
        dt: float = 5e-15,
        t0: Optional[Tuple[float, float]] = None,
        f0: float = 750e12,
    ):
        """Bin two different spectrograms, e.g. from different spectrometers, onto the time time/frequency grid.

        Args:
            other: the other spectrogram
            stitching_band (Tuple[float, float]): the lower and upper frequency of the band where the two spectrometers should have equivalent response (hopefully there is one)
            dim (int): size of each size of the resulting square data
            dt (float): time step of the data
            t0: (Optional[Tuple[float, float]): time-zero of the data (this, and other). If not specified, will be calculated by the first moment of the time-distribution of the signal
            f0: (float): central frequency of the binned array

        """
        t0_self = None
        t0_other = None

        if t0 is not None:
            t0_self = t0[0]
            t0_other = t0[1]

        binned_self = self.to_binned(dim, dt, t0_self, f0)
        binned_other = other.to_binned(dim, dt, t0_other, f0)
        freq = binned_self.freq

        # add more logic here to combine the spectrograms
        stitching_band_integral_self = np.sum(
            binned_self.data[
                ((freq > stitching_band[0]) & (freq < stitching_band[1])), :
            ][:]
        )
        stitching_band_integral_other = np.sum(
            binned_other.data[
                ((freq > stitching_band[0]) & (freq < stitching_band[1])), :
            ][:]
        )
        weights_self = np.zeros(binned_self.freq.shape, dtype=float)
        weights_other = np.zeros(binned_other.freq.shape, dtype=float)
        other_multiplier = stitching_band_integral_self / stitching_band_integral_other
        for i in range(len(freq)):
            sum_self = np.sum(binned_self.data[i, :])
            sum_other = other_multiplier * np.sum(binned_other.data[i, :])
            total = sum_self + sum_other
            if total > 0.0:
                weight_self = sum_self / total
                weight_other = other_multiplier * sum_other / total
                weights_self[i] = weight_self
                weights_other[i] = weight_other

        return Spectrogram(
            data=weights_self[:, np.newaxis] * binned_self.data
            + weights_other[:, np.newaxis] * binned_other.data,
            time=binned_self.time,
            freq=binned_self.freq,
        )

    def to_binned(
        self,
        dim: int = 64,
        dt: float = 5e-15,
        t0: Optional[float] = None,
        f0: float = 750e12,
    ):
        """Bin a spectrogram to a FFT-appropriate shape.

        Args:
            dim (int): size of each size of the resulting square data
            dt (float): time step of the data
            t0: (Optional[float]): time-zero of the data. If not specified, will be calculated by the first moment of the time-distribution of the signal
            f0: (float): central frequency of the binned array

        Returns:
            Spectrogram: the binned spectrogram

        """
        _t = np.array(range(dim)) * dt
        _t -= np.mean(_t)
        _f = np.fft.fftshift(np.fft.fftfreq(dim, d=dt) + f0)
        binned_data = np.zeros((dim, self.time.shape[0]), dtype=float)
        for _i in range(self.time.shape[0]):
            binned_data[:, _i] = interpolate(
                _f, self.freq, np.array(self.data[:, _i]), neighbors=2
            )
        binned_data /= np.max(binned_data[:])
        if t0 is None:
            ac = np.sum(binned_data, axis=0)
            t0 = np.sum(ac * self.time) / np.sum(ac)
        binned_data_square = np.zeros((dim, dim), dtype=float)
        for _i in range(dim):
            binned_data_square[_i, :] = interpolate(
                _t, self.time - t0, np.array(binned_data[_i, :]), neighbors=2
            )
        return Spectrogram(data=binned_data_square, time=_t, freq=_f)

    def to_bin_pipeline_result(self, settings: FrogBinSettings):
        """Apply a FrogBinSettings dataclass to the spectrogram.

        Args:
            settings (FrogBinSettings): the dataclass to apply

        """
        if settings.median_binning:
            method = "media"
        else:
            method = "mean"

        t0 = settings.t0
        if settings.auto_t0:
            t0 = None

        def maybe_correct_chirp(instance, apply: bool):
            return instance.to_removed_spatial_chirp() if apply else instance

        return (
            maybe_correct_chirp(self, settings.spatial_chirp_correction)
            .to_block_binned(settings.freq_binning, settings.time_binning, method)
            .to_binned(dim=settings.size, dt=settings.dt, f0=settings.f0, t0=t0)
            .to_per_frequency_dc_removed(extra_offset=settings.dc_offset)
        )

    def plot(self, ax: Optional[Axes] = None, take_sqrt: bool = True):
        """Plot the spectrogram.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes
            take_sqrt (bool): plot the square root of the data

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        if take_sqrt:
            a = ax.pcolormesh(
                1e15 * self.time,
                1e-12 * self.freq,
                np.sqrt(self.data / np.max(self.data[:])),
                rasterized=True,
            )
        else:
            a = ax.pcolormesh(
                1e15 * self.time,
                1e-12 * self.freq,
                self.data / np.max(self.data[:]),
                rasterized=True,
            )
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Frequency (THz)")
        plt.colorbar(a)
        return fig

    def plot_log(self, ax: Optional[Axes] = None):
        """Plot the spectrogram.

        Args:
            ax: optionally plot onto a pre-existing matplotlib Axes

        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        logdata = np.array(self.data)
        logdata[self.data > 0.0] = np.log(self.data[self.data > 0.0])
        logdata[self.data < 0.0] = 0.0
        a = ax.pcolormesh(
            1e15 * self.time,
            1e-12 * self.freq,
            logdata,
            rasterized=True,
            vmin=-11,
            vmax=0,
        )
        ax.set_xlabel("Time (fs)")
        ax.set_ylabel("Frequency (THz)")
        ax.grid(True, lw=1)
        plt.colorbar(a)
        return fig



"""Class and function decorators."""

from dataclasses import is_dataclass

import numpy as np
import yaml


def add_method(cls, name: str):
    """Adds the decorated function as a member function of a class. The first argument of the function should be 'self'.

    Args:
        cls: The class type
        name: What the method will be called

    """

    def decorator(func):
        setattr(cls, name, func)
        return func

    return decorator


def yaml_io(cls):
    """Adds functions to save and load the dataclass as yaml."""

    def from_dict(cls, data: dict):
        """Takes a dict and makes an instance of the class.

        Args:
            cls: the class (hidden)
            data (dict): the result of a call of .to_dict on the class

        """

        def handle_complex_array(serialized_array) -> np.ndarray:
            """Helper function to deserialize numpy arrays, handling complex types."""
            if isinstance(serialized_array, list) and all(
                isinstance(item, dict) and "re" in item and "im" in item
                for item in serialized_array
            ):
                return np.array(
                    [complex(item["re"], item["im"]) for item in serialized_array],
                    dtype=np.complex128,
                )
            return np.array(serialized_array)

        loaded_data = {}
        for field_name, field_type in cls.__annotations__.items():
            if field_type is np.ndarray:
                loaded_data[field_name] = handle_complex_array(data[field_name])
            elif is_dataclass(field_type):
                loaded_data[field_name] = field_type.from_dict(data[field_name])
            else:
                loaded_data[field_name] = data[field_name]
        return cls(**loaded_data)

    def load_yaml(cls, filename: str):
        """Load from a yaml file.

        Args:
            cls: the class (hidden)
            filename (str): path to the file

        """
        with open(filename, "r") as file:
            data = yaml.load(file, yaml.SafeLoader)
            return cls.from_dict(data)

    def load_yaml_bytestream(cls, stream):
        data = yaml.load(stream, yaml.SafeLoader)
        return cls.from_dict(data)

    def save_yaml(instance, filename: str):
        """Save to a yaml file.

        Args:
            instance: the class (hidden)
            filename (str): path to the file

        """
        data_dict = instance.to_dict()
        with open(filename, "w") as file:
            yaml.dump(data_dict, file)

    def to_dict(instance):
        """Serialize the class into a dict."""
        data_dict = {}
        for field_name, field_type in instance.__annotations__.items():
            field_value = getattr(instance, field_name)
            if field_type is np.ndarray:
                if field_value.dtype == np.complex128:
                    data_dict[field_name] = [
                        {"re": num.real, "im": num.imag} for num in field_value.tolist()
                    ]
                else:
                    data_dict[field_name] = field_value.tolist()
            elif is_dataclass(field_type):
                data_dict[field_name] = field_value.to_dict()
            elif field_type is np.float64 or field_type is float:
                data_dict[field_name] = float(field_value)
            else:
                data_dict[field_name] = field_value
        return data_dict

    cls.from_dict = classmethod(from_dict)
    cls.load_yaml = classmethod(load_yaml)
    cls.load_yaml_bytestream = classmethod(load_yaml_bytestream)
    cls.to_dict = to_dict
    cls.save_yaml = save_yaml
    return cls



import marimo

__generated_with = "0.15.2"
app = marimo.App(width="medium")


@app.cell
async def _():
    import marimo as mo
    # check if running in a browser, install attoworld from local copy
    import sys
    is_in_web_notebook = sys.platform == "emscripten"
    if is_in_web_notebook:
        import micropip
        import os
        await micropip.install("https://nickkarpowicz.github.io/wheels/attoworld-2025.0.41-cp312-cp312-emscripten_3_1_58_wasm32.whl")

        import base64
        import zipfile

        def create_download_link(data, filename, mime_type="text/plain"):
            encoded_data = base64.b64encode(data).decode("utf-8")
            data_uri = f"data:{mime_type};base64,{encoded_data}"
            html = f'<a href="{data_uri}" download="{filename}">Download {filename}</a>'
            return mo.Html(html)

        def display_download_link_from_file(
            path, output_name, mime_type="text/plain"
        ):
            with open(path, "rb") as _file:
                mo.output.append(
                    create_download_link(
                        data=_file.read(),
                        filename=output_name,
                        mime_type=mime_type,
                    )
                )
    else:
        import tkinter as tk
        from tkinter import filedialog

        root = tk.Tk()
        root.withdraw()

    import attoworld as aw
    import numpy as np
    import pathlib

    aw.plot.set_style("nick_dark")
    return (
        aw,
        display_download_link_from_file,
        filedialog,
        is_in_web_notebook,
        mo,
        np,
        pathlib,
        zipfile,
    )


@app.cell
def _(mo):
    mo.md(
        r"""
    # FROG reconstruction

    ---
    #### Select your FROG file:
    """
    )
    return


@app.cell
def _(mo):
    file_browser = mo.ui.file(filetypes=[".dwc"], label="Select .dwc file")
    file_browser
    return (file_browser,)


@app.cell
def _(aw, mo):
    calibration_selector = mo.ui.dropdown(options=[e.value for e in aw.spectrum.CalibrationData],label="Calibration:")
    calibration_selector
    return (calibration_selector,)


@app.cell
def _(mo):
    mode_selector = mo.ui.dropdown(options=["SHG", "THG", "Kerr", "XFROG", "BlindFROG"], label="FROG type:", value="SHG")
    mode_selector
    return (mode_selector,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ---
    #### Optional spectral constraint:
    """
    )
    return


@app.cell
def _(mo):
    spectral_constraint_file = mo.ui.file(label="Spectral contstraint file")
    spectral_constraint_file
    return (spectral_constraint_file,)


@app.cell
def _(mo, spectral_constraint_file):
    spectral_constraint_format = mo.ui.dropdown(options=["Columns", "Text with headers"], value="Text with headers")
    spectral_constraint_data = spectral_constraint_file.contents()
    if spectral_constraint_data is not None:
        mo.output.append(spectral_constraint_format)
    return spectral_constraint_data, spectral_constraint_format


@app.cell
def _(mo, spectral_constraint_data, spectral_constraint_format):
    spectral_constraint_wavelength_header = mo.ui.text(label="Wavelength column key:", value="wavelength (nm)")
    spectral_constraint_wavelength_multiplier = mo.ui.number(label="Wavelength multiplier:", value=1e9)
    spectral_constraint_intensity_header = mo.ui.text(label="Intensity column key:", value="intensity (a.u.)")
    spectral_constraint_skip_lines = mo.ui.number(value=0, label="Header lines:")
    if spectral_constraint_data is not None:
        if spectral_constraint_format.value == "Text with headers":
            mo.output.append(spectral_constraint_wavelength_header)
            mo.output.append(spectral_constraint_wavelength_multiplier)
            mo.output.append(spectral_constraint_intensity_header)
        if spectral_constraint_format.value == "Columns":
            mo.output.append(spectral_constraint_skip_lines)
    return (
        spectral_constraint_intensity_header,
        spectral_constraint_skip_lines,
        spectral_constraint_wavelength_header,
        spectral_constraint_wavelength_multiplier,
    )


@app.cell
def _(
    aw,
    mo,
    spectral_constraint_data,
    spectral_constraint_format,
    spectral_constraint_intensity_header,
    spectral_constraint_skip_lines,
    spectral_constraint_wavelength_header,
    spectral_constraint_wavelength_multiplier,
):
    spectral_constraint = None
    if spectral_constraint_data is not None:
        match spectral_constraint_format.value:
            case "Columns":
                spectral_constraint = aw.data.load_mean_spectrum_from_scarab(
                    spectral_constraint_data.decode("utf-8"), is_bytes=True, header_size=spectral_constraint_skip_lines.value
                )
            case "Text with headers":
                spectral_constraint = aw.data.load_spectrum_from_text(
                    filename=spectral_constraint_data,
                    wavelength_multiplier=1.0
                    / spectral_constraint_wavelength_multiplier.value,
                    wavelength_field=spectral_constraint_wavelength_header.value,
                    spectrum_field=spectral_constraint_intensity_header.value,
                )
        mo.output.append(mo.md("### Loaded spectral constraint:"))
        spectral_constraint.plot_with_group_delay()
        aw.plot.showmo()
    return (spectral_constraint,)


@app.cell
def _(mo, mode_selector):
    xfrog_reference_file = mo.ui.file(filetypes=[".yml",".dat"], label="Select reference file")
    xfrog_time_reverse_checkbox = mo.ui.checkbox(label="Reverse time")
    if(mode_selector.value == "XFROG"):
        mo.output.append(mo.md("---"))
        mo.output.append(mo.md("#### XFROG Reference:"))
        mo.output.append(xfrog_reference_file)
        mo.output.append(xfrog_time_reverse_checkbox)
    return xfrog_reference_file, xfrog_time_reverse_checkbox


@app.cell
def _(
    aw,
    mo,
    mode_selector,
    np,
    pathlib,
    xfrog_reference_file,
    xfrog_time_reverse_checkbox,
):
    if((mode_selector.value == "XFROG") and (xfrog_reference_file.name() is not None)):
        mo.output.append(mo.md("### Loaded reference:"))
        _type = pathlib.Path(xfrog_reference_file.name()).suffix
        if _type == ".yml":
            xfrog_reference = aw.data.FrogData.load_yaml_bytestream(xfrog_reference_file.contents())
            if xfrog_time_reverse_checkbox.value:
                xfrog_reference.raw_reconstruction = np.flipud(xfrog_reference.raw_reconstruction)
                xfrog_reference.pulse.wave = np.flipud(xfrog_reference.pulse.wave)
                xfrog_reference.spectrum.spectrum = np.conj(xfrog_reference.spectrum.spectrum)
            xfrog_reference.plot_all(figsize=(9,6))
            aw.plot.showmo()
    else:
        xfrog_reference=None
    return (xfrog_reference,)


@app.cell
def _(aw, calibration_selector, file_browser):
    _path = file_browser.contents()
    if _path is not None:
        input_data = aw.data.read_dwc(file_or_path=_path, is_buffer=True)
        if calibration_selector.value is not None:
            calibration = aw.data.SpectrometerCalibration.from_npz(
                aw.spectrum.get_calibration_path() / calibration_selector.value
            )
            input_data = calibration.apply_to_spectrogram(input_data)
    else:
        input_data = None
    return (input_data,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ---
    #### Bin onto evenly-spaced space/time grid:
    """
    )
    return


@app.cell
def _(mo):
    bin_loaded_file = mo.ui.file(filetypes=[".yml"], label="Load settings from .yml")
    bin_loaded_file
    return (bin_loaded_file,)


@app.cell
def _(aw, bin_loaded_file):
    _contents = bin_loaded_file.contents()
    if _contents is not None:
        loaded_settings = aw.data.FrogBinSettings.load_yaml_bytestream(_contents)
    else:
        loaded_settings = aw.data.FrogBinSettings(
            size=96,
            dt=3e-15,
            t0=0.0,
            auto_t0=True,
            f0=740e12,
            dc_offset=0.0002,
            freq_binning=1,
            time_binning=1,
            median_binning=False,
            spatial_chirp_correction=False,
        )
    return (loaded_settings,)


@app.cell
def _(is_in_web_notebook, loaded_settings, mo):
    bin_size = mo.ui.number(label="size", value=loaded_settings.size, step=2)
    bin_dt = mo.ui.number(label="dt (fs)", value=loaded_settings.dt*1e15, step=0.1)
    bin_t0 = mo.ui.number(label="t0 (fs)", value=loaded_settings.t0 * 1e-15, step=0.1)
    bin_t0_auto = mo.ui.checkbox(label="Auto time centering", value=loaded_settings.auto_t0)
    bin_f0 = mo.ui.number(label="f0 (THz)", value=loaded_settings.f0*1e-12, step=0.1)
    bin_offset = mo.ui.number(label="dark noise level", value=loaded_settings.dc_offset, step=1e-5)
    bin_fblock = mo.ui.number(label="freq block avg.", start=1, value=loaded_settings.freq_binning, step=1)
    bin_tblock = mo.ui.number(label="time block avg.", start=1, value=loaded_settings.time_binning, step=1)
    bin_median = mo.ui.checkbox(label="median blocking", value=loaded_settings.median_binning)
    bin_spatial_chirp_correction = mo.ui.checkbox(label="correct spatial chirp", value=loaded_settings.spatial_chirp_correction)
    mo.output.append(bin_size)
    mo.output.append(bin_dt)
    mo.output.append(bin_t0)
    mo.output.append(bin_t0_auto)
    mo.output.append(bin_f0)
    mo.output.append(bin_offset)
    mo.output.append(bin_fblock)
    mo.output.append(bin_tblock)
    mo.output.append(bin_median)
    mo.output.append(bin_spatial_chirp_correction)

    if not is_in_web_notebook:
        bin_save_button = mo.ui.run_button(label="Save settings")
        mo.output.append(bin_save_button)
    return (
        bin_dt,
        bin_f0,
        bin_fblock,
        bin_median,
        bin_offset,
        bin_save_button,
        bin_size,
        bin_spatial_chirp_correction,
        bin_t0,
        bin_t0_auto,
        bin_tblock,
    )


@app.cell
def _(
    aw,
    bin_dt,
    bin_f0,
    bin_fblock,
    bin_median,
    bin_offset,
    bin_size,
    bin_spatial_chirp_correction,
    bin_t0,
    bin_t0_auto,
    bin_tblock,
):
    if bin_t0_auto.value:
        _t0 = None
    else:
        _t0 = bin_t0.value * 1e-15
    bin_settings = aw.data.FrogBinSettings(
        size=int(bin_size.value),
        dt=bin_dt.value * 1e-15,
        t0=bin_t0.value * 1e-15,
        auto_t0=bin_t0_auto.value,
        f0=bin_f0.value * 1e12,
        dc_offset=bin_offset.value,
        time_binning=int(bin_tblock.value),
        freq_binning=int(bin_fblock.value),
        median_binning=bool(bin_median.value),
        spatial_chirp_correction=bool(bin_spatial_chirp_correction.value),
    )
    return (bin_settings,)


@app.cell
def _(
    aw,
    bin_settings,
    display_download_link_from_file,
    input_data,
    is_in_web_notebook,
):
    # if not bin_live.value:
    #     mo.stop(not bin_button.value)
    if input_data is not None:
        frog_data = input_data.to_bin_pipeline_result(bin_settings)
        frog_data.plot_log()
        aw.plot.showmo()

        if is_in_web_notebook:
            bin_settings.save_yaml("bin_settings.yml")
            display_download_link_from_file(
                path="bin_settings.yml",
                output_name="bin_settings.yml",
                mime_type="text/yaml",
            )

    else:
        frog_data = None
    return (frog_data,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ---
    #### Run the reconstruction routine:
    """
    )
    return


@app.cell
def _(is_in_web_notebook, mo):
    recon_trials = mo.ui.number(value=8, label="Initial guesses")
    recon_trial_length = mo.ui.number(value=64, label="Trial iterations")
    recon_followups = mo.ui.number(value=512, label="Finishing iterations")
    reconstruct_button = mo.ui.run_button(label="reconstruct")
    save_button = mo.ui.run_button(label="save")
    save_plot_button = mo.ui.run_button(label="save plot")
    mo.output.append(recon_trials)
    mo.output.append(recon_trial_length)
    mo.output.append(recon_followups)
    mo.output.append(reconstruct_button)

    if not is_in_web_notebook:
        mo.output.append(save_button)
        mo.output.append(save_plot_button)
    return (
        recon_followups,
        recon_trial_length,
        recon_trials,
        reconstruct_button,
        save_button,
        save_plot_button,
    )


@app.cell
def _(is_in_web_notebook, mo):
    if is_in_web_notebook:
        file_base = mo.ui.text(value="output", label="Output name")
        mo.output.append(file_base)
    return (file_base,)


@app.cell
def _(
    aw,
    frog_data,
    mo,
    mode_selector,
    recon_followups,
    recon_trial_length,
    recon_trials,
    reconstruct_button,
    spectral_constraint,
    xfrog_reference,
):
    mo.stop(not reconstruct_button.value)
    if frog_data is not None:
        match mode_selector.value:
            case "SHG":
                frog_type = aw.attoworld_rs.FrogType.Shg
            case "THG":
                frog_type = aw.attoworld_rs.FrogType.Thg
            case "Kerr":
                frog_type = aw.attoworld_rs.FrogType.Kerr
            case "XFROG":
                frog_type = aw.attoworld_rs.FrogType.Xfrog
            case "BlindFROG":
                frog_type = aw.attoworld_rs.FrogType.Blindfrog
        result, gate_result = aw.wave.reconstruct_frog(
            measurement=frog_data,
            repeats=int(recon_trials.value),
            test_iterations=int(recon_trial_length.value),
            polish_iterations=int(recon_followups.value),
            frog_type=frog_type,
            spectrum=spectral_constraint,
            xfrog_gate = xfrog_reference
        )
    else:
        result = None
    return (result,)


@app.cell
def _(
    aw,
    display_download_link_from_file,
    file_base,
    is_in_web_notebook,
    mo,
    mode_selector,
    result,
    result_gate,
    zipfile,
):
    if result is not None:
        plot = result.plot_all(
            figsize=(9, 6),
            wavelength_autoscale=1e-3
        )
        aw.plot.showmo()
        if mode_selector.value == "BlindFROG":
            mo.output.append(mo.md("### Gate"))
            plot_gate = result_gate.plot_all(
                figsize=(9, 6),
                wavelength_autoscale=1e-3
            )
            aw.plot.showmo()

        if is_in_web_notebook:
            plot.savefig("temp.svg")
            display_download_link_from_file(
                "temp.svg", output_name=f"{file_base.value}.svg"
            )

            result.save(file_base.value)
            result.save_yaml(f"{file_base.value}.yml")
            with zipfile.ZipFile(f"{file_base.value}.zip", "w") as zip:
                zip.write(f"{file_base.value}.A.dat")
                zip.write(f"{file_base.value}.Arecon.dat")
                zip.write(f"{file_base.value}.Ek.dat")
                zip.write(f"{file_base.value}.Speck.dat")
                zip.write(f"{file_base.value}.yml")
            display_download_link_from_file(
                f"{file_base.value}.zip",
                output_name=f"{file_base.value}.zip",
                mime_type="application/zip",
            )
            display_download_link_from_file(f"{file_base.value}.yml",output_name=f"{file_base.value}.yml",mime_type="text/yaml")
    return (plot,)


@app.cell
def _(filedialog, is_in_web_notebook, mo, result, save_button):
    mo.stop(not save_button.value)
    if not is_in_web_notebook:
        _file_path = filedialog.asksaveasfilename(
            title="Save File", filetypes=[("All Files", "*.*")]
        )

        if (_file_path is not None) and (result is not None) and (_file_path != ""):
            result.save(_file_path)
            result.save_yaml(_file_path + ".yml")
    return


@app.cell
def _(filedialog, is_in_web_notebook, mo, plot, result, save_plot_button):
    mo.stop(not save_plot_button.value)
    if not is_in_web_notebook:
        _file_path = filedialog.asksaveasfilename(
            title="Save File", filetypes=[("SVG files", "*.svg"), ("PDF files", "*.pdf")]
        )

        if _file_path is not None and result is not None:
            plot.savefig(_file_path)
    return


@app.cell
def _(bin_save_button, bin_settings, filedialog, is_in_web_notebook, mo):
    if not is_in_web_notebook:
        mo.stop(not bin_save_button.value)
        _file_path = filedialog.asksaveasfilename(
            title="Save File", filetypes=[("YAML files", "*.yml")]
        )

        if _file_path is not None and bin_settings is not None:
            bin_settings.save_yaml(_file_path)
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



"""A collection of numerical tools."""

from ..attoworld_rs import (
    derivative,
    derivative_periodic,
    find_first_intercept,
    find_last_intercept,
    find_maximum_location,
    fornberg_stencil,
    fwhm,
    interpolate,
)
from .atomic_units import AtomicUnits
from .numeric import block_binning_1d, block_binning_2d

__all__ = [
    "fornberg_stencil",
    "find_maximum_location",
    "fwhm",
    "find_first_intercept",
    "find_last_intercept",
    "derivative",
    "derivative_periodic",
    "interpolate",
    "AtomicUnits",
    "block_binning_1d",
    "block_binning_2d",
]



"""Handle the results from a Luna simulation."""

import h5py
import matplotlib.pyplot as plt
import numpy as np
from scipy import constants

from ..plot import Char


def check_equal_length(*arg):
    """Make ture the size matches."""
    n = len(arg[0])
    for v in arg:
        if len(v) != n:
            print(v)
            print("Error: vector size mismatch")
            raise Exception("Vector size mismatch")


def fourier_transform(TimeV, FieldV):
    """Apply forward transform."""
    freq = np.fft.fftfreq(TimeV.size, d=TimeV[1] - TimeV[0])
    fft = np.fft.fft(FieldV)
    return freq, fft


def inverse_fourier_transform(freq, fullSpectrum):
    """Apply inverse transform."""
    timeV = np.fft.fftfreq(len(freq), freq[1] - freq[0])
    if len(timeV) % 2 == 0:
        timeV = np.concatenate(
            (timeV[int(len(timeV) / 2) :], timeV[: int(len(timeV) / 2)])
        )
    else:
        timeV = np.concatenate(
            (timeV[int((len(timeV) + 1) / 2) :], timeV[: int((len(timeV) + 1) / 2)])
        )
    fieldV = np.fft.ifft(fullSpectrum)

    return timeV, fieldV


class LunaResult:
    """Loads and handles the Luna simulation result.

    The result must be in the HDF5 format using the saving option in the Luna.Interface.prop_capillary() function [filepath="..."].
    As opposed to most of the analysis routines here, units are in SI!

    Attributes:
        z: position along the fiber (for the field data)
        fieldFT: complex FFT of the field (positive freq axis); shape: (n_z, n_modes, n_freq) [or (n_z, n_freq) after mode selection/averaging]
        omega: angular frequency axis for the FFT field; shape: (n_z, n_modes, n_freq) [or (n_z, n_freq) after mode selection/averaging]

        stats_z: position along the fiber (for stats data)
        stats_density: particle density along the fiber (m^-3)
        stats_electrondensity: electron density along the fiber (m^-3)
        stats_pressure: gas pressure profile along the fiber (bar)
        stats_energy: pulse energy along the fiber (J)
        stats_peakpower: peak power of the pulse along the fiber (W)
        stats_peakintensity: peak intensity of the pulse along the fiber
        stats_peak_ionization_rate: peak ionization rate along the fiber

    """

    def __init__(self, filename):
        """Constructor of class LunaResult.

        Args:
            filename: saved result, file path

        """
        self.filename = filename
        self.fieldFT = None
        self.omega = None
        self.z = None

        self.stats_z = None
        self.stats_energy = None
        self.stats_electrondensity = None
        self.stats_density = None
        self.stats_pressure = None
        self.stats_peakpower = None
        self.stats_peakintensity = None
        self.stats_zdw = None
        self.stats_peak_ionization_rate = None

        self.open_Luna_result(filename)

    def open_Luna_result(self, filename):
        """Opens the Luna result file and loads the data."""
        with h5py.File(filename, "r") as data:
            # FIELD DATA
            self.fieldFT = np.array(data["Eω"])
            self.omega = np.array(data["grid"]["ω"])
            self.z = np.array(data["z"])

            # STATS
            self.stats_z = np.array(data["stats"]["z"])
            self.stats_energy = np.array(data["stats"]["energy"])
            self.stats_electrondensity = np.array(data["stats"]["electrondensity"])
            self.stats_density = np.array(data["stats"]["density"])
            self.stats_pressure = np.array(data["stats"]["pressure"])
            self.stats_peakpower = np.array(data["stats"]["peakpower"])
            self.stats_peakintensity = np.array(data["stats"]["peakintensity"])
            self.stats_zdw = np.array(data["stats"]["zdw"])
            self.stats_peak_ionization_rate = np.array(
                data["stats"]["peak_ionisation_rate"]
            )

    def average_modes(self):
        """Averages the propagation modes in the Luna result file."""
        if len(self.fieldFT.shape) == 3:
            self.fieldFT = np.mean(self.fieldFT, axis=1)
            self.stats_zdw = None
            self.stats_peakpower = None
            self.stats_energy = np.sum(self.stats_energy, axis=1)

    def select_mode(self, mode: int):
        """Select the fiber mode."""
        if len(self.fieldFT.shape) < 3:
            print("WARNING: No mode to select")
        elif mode >= self.fieldFT.shape[1] or mode < 0:
            print("WARNING: mode ", mode, " is out of range")
        else:
            self.fieldFT = self.fieldFT[:, mode, :]
            self.stats_zdw = self.stats_zdw[:, mode]
            self.stats_peakpower = self.stats_peakpower[:, mode]
            self.stats_energy = self.stats_energy[:, mode]

    def get_time_field(self, position=None):
        """Get the electric field in time from the Luna result file. If no mode was previously selected, the method computes the average of all modes.
        Therefore, after calling get_time_field(), mode selection is not possible any more.

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            timeV (numpy.ndarray): time axis in seconds
            fieldV (numpy.ndarray): electric field in V/m

        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        check_equal_length(self.fieldFT[index], self.omega)
        fieldFFT = np.concatenate(
            (self.fieldFT[index, :], np.conjugate(self.fieldFT[index, :][::-1]) * 0)
        )
        freq = np.concatenate((self.omega, -self.omega[::-1])) / 2 / np.pi
        timeV, fieldV = inverse_fourier_transform(freq, fieldFFT)
        return timeV, np.real(fieldV)

    def get_wavelength_spectrum(self, position=None):
        """Get the spectrum from the Luna result file (|FFT|^2 * (2 * pi * c / λ^2)).

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            wvl (numpy.ndarray): wavelength axis in m
            wvlSpectrum (numpy.ndarray): electric field spectrum in V/m

        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        wvl = 2 * np.pi * constants.speed_of_light / self.omega[::-1]
        wvlSpectrum = np.abs(self.fieldFT[index, ::-1]) ** 2 * (
            2 * np.pi * constants.speed_of_light / wvl**2
        )
        return wvl, wvlSpectrum

    def get_spectral_phase(self, position=None):
        """Get the spectral phase from the Luna result file.

        Args:
            position (float): position along the fiber in m. If None, the end of the fiber is used.

        Returns:
            wvl (numpy.ndarray): wavelength axis in m
            phase (numpy.ndarray): spectral phase in rad

        """
        self.average_modes()
        if position is None:
            position = self.z[-1]
        index = np.argmin(np.abs(self.z - position))
        if position > np.max(self.z) or position < np.min(self.z):
            print("WARNING: position ", position, "m is out of range")
        wvl = 2 * np.pi * constants.speed_of_light / self.omega[::-1]
        phase = np.angle(self.fieldFT[index, ::-1])
        return wvl, phase

    def plot_stats(self):
        """Plots the 'stats_' attribute of the simulation stored in the present object."""
        fig, axs = plt.subplots(2, 2, figsize=[7, 5])
        axs[0, 0].plot(self.stats_z, self.stats_pressure)
        axs[0, 0].set_xlabel("z (m)")
        axs[0, 0].set_ylabel("gas pressure (bar)")
        ax2 = axs[0, 0].twinx()
        ax2.plot(self.stats_z, self.stats_density, color="r")
        ax2.set_ylabel("gas particle density ($m^{-3}$)", color="r")
        ax2.tick_params(axis="y", colors="r")
        ax2.yaxis.label.set_color("r")
        axs[0, 1].plot(self.stats_z, self.stats_electrondensity)
        axs[0, 1].set_xlabel("z (m)")
        axs[0, 1].set_ylabel("electron density ($m^{-3}$)")
        ax2 = axs[0, 1].twinx()
        if len(self.stats_energy.shape) == 2:
            ax2.plot(self.stats_z, np.sum(self.stats_energy, axis=1) * 1e6, color="r")
        else:
            ax2.plot(self.stats_z, self.stats_energy * 1e6, color="r")
        ax2.set_ylabel(f"pulse energy ({Char.mu}J)", color="r")
        ax2.tick_params(axis="y", colors="r")
        ax2.yaxis.label.set_color("r")
        if self.stats_peakpower is not None:
            if len(self.stats_peakpower.shape) == 2:
                axs[1, 0].plot(self.stats_z, self.stats_peakpower[:, 0])
                axs[1, 0].set_xlabel("z (m)")
                axs[1, 0].set_ylabel("peak power (W)")
            elif len(self.stats_peakpower.shape) == 1:
                axs[1, 0].plot(self.stats_z, self.stats_peakpower)
                axs[1, 0].set_xlabel("z (m)")
                axs[1, 0].set_ylabel("peak power (W)")
        axs[1, 1].plot(self.stats_z, self.stats_peak_ionization_rate)
        axs[1, 1].set_xlabel("z (m)")
        axs[1, 1].set_ylabel("peak ionization rate")
        plt.tight_layout()

        return fig



"""Contains functions related to the processing of spectra."""

from ..personal.marco import (
    load_calibration_data,
    read_spectrometer_excel,
    read_spectrum_ocean_optics,
)
from .calibration_data import (
    CalibrationData,
    CalibrationLampReferences,
    get_calibration_path,
)
from .spectrum import (
    frequency_to_wavelength,
    transform_limited_pulse_from_spectrometer,
    wavelength_to_frequency,
)

__all__ = [
    "CalibrationData",
    "CalibrationLampReferences",
    "get_calibration_path",
    "wavelength_to_frequency",
    "frequency_to_wavelength",
    "transform_limited_pulse_from_spectrometer",
    "load_calibration_data",
    "read_spectrometer_excel",
    "read_spectrum_ocean_optics",
]



"""Operations for converting the data containters into each other."""

from typing import Optional

import numpy as np
import scipy.signal as sig
from scipy import constants

from ..numeric import (
    find_maximum_location,
)
from .complex_envelope import ComplexEnvelope
from .complex_spectrum import ComplexSpectrum
from .decorators import add_method
from .intensity_spectrum import IntensitySpectrum
from .waveform import Waveform


@add_method(Waveform, "to_complex_spectrum")
def waveform_to_complex_spectrum(self, padding_factor: int = 1):
    """Converts to a ComplexSpectrum class.

    Args:
        self: the class
        padding_factor (int): factor by which to expand the temporal length in the FFT, giving a smoother spectrum

    """
    if self.is_uniformly_spaced:
        new_spectrum = np.fft.rfft(
            self.wave, n=self.wave.shape[0] * padding_factor, axis=0
        )
        new_freq = np.fft.rfftfreq(self.wave.shape[0] * padding_factor, d=self.dt)
    else:
        uniform_self = self.to_uniformly_spaced()
        new_spectrum = np.fft.rfft(
            uniform_self.wave, n=uniform_self.wave.shape[0] * padding_factor, axis=0
        )
        new_freq = np.fft.rfftfreq(
            uniform_self.wave.shape[0] * padding_factor, d=uniform_self.dt
        )

    return ComplexSpectrum(spectrum=new_spectrum, freq=new_freq)


@add_method(Waveform, "to_intensity_spectrum")
def waveform_to_intensity_spectrum(
    self, wavelength_scaled: bool = True, padding_factor: int = 1
):
    """Converts to an intensity spectrum.

    Args:
        self: the class
        wavelength_scaled (bool): Correct the spectral intensities for plotting on a wavelength scale
        padding_factor (int): the factor by which the length will be multiplied

    """
    return self.to_complex_spectrum(padding_factor).to_intensity_spectrum(
        wavelength_scaled
    )


@add_method(Waveform, "to_time_derivative")
def waveform_to_time_derivative(self):
    """Return the time-derivative of the waveform."""
    return self.to_complex_spectrum().to_time_derivative().to_waveform()


@add_method(Waveform, "to_complex_envelope")
def waveform_to_complex_envelope(self, f0: float = 0.0):
    """Return a ComplexEnvelope class corresponding to the waveform.

    Args:
        self: the class
        f0 (float): central frequency to use when constructing the envelope. E.g. oscillation at this frequency will be cancelled.

    """
    uniform_self = self.to_uniformly_spaced()
    analytic = np.array(
        sig.hilbert(uniform_self.wave)
        * np.exp(-1j * 2 * np.pi * f0 * uniform_self.time)
    )
    return ComplexEnvelope(
        envelope=analytic,
        time=uniform_self.time,
        dt=uniform_self.dt,
        carrier_frequency=f0,
    )


@add_method(ComplexSpectrum, "to_waveform")
def complexspectrum_to_waveform(self):
    """Create a Waveform based on this complex spectrum."""
    wave = np.fft.irfft(self.spectrum, axis=0)
    dt = 0.5 / (self.freq[-1] - self.freq[0])
    time = dt * np.array(range(wave.shape[0]))
    return Waveform(wave=wave, time=time, dt=dt, is_uniformly_spaced=True)


@add_method(ComplexSpectrum, "to_centered_waveform")
def complexspectrum_to_centered_waveform(self):
    """Create a Waveform based on this complex spectrum and center it in the time window."""
    wave = np.fft.irfft(self.spectrum, axis=0)
    dt = 0.5 / (self.freq[-1] - self.freq[0])
    time = dt * np.array(range(wave.shape[0]))
    max_ind, max_val = find_maximum_location(np.abs(np.array(sig.hilbert(wave))))
    max_loc = dt * max_ind - 0.5 * time[-1]
    wave = np.fft.irfft(
        np.exp(-1j * self.freq * 2 * np.pi * max_loc) * self.spectrum, axis=0
    )
    return Waveform(wave=wave, time=time, dt=dt, is_uniformly_spaced=True)


@add_method(ComplexSpectrum, "to_intensity_spectrum")
def complexspectrum_to_intensity_spectrum(self, wavelength_scaled: bool = True):
    """Create an IntensitySpectrum based on the current ComplexSpectrum.

    Args:
        self: the class
        wavelength_scaled (bool): Apply the wavelength^-2 Jakobian such to correspond to W/nm spectrum

    """
    new_spectrum = np.array(np.abs(self.spectrum[self.freq > 0.0]) ** 2)
    new_freq = np.array(self.freq[self.freq > 0.0])
    new_wavelength = constants.speed_of_light / new_freq
    if wavelength_scaled:
        new_spectrum /= new_wavelength**2
    return IntensitySpectrum(
        spectrum=new_spectrum,
        phase=np.array(np.angle(self.spectrum[self.freq > 0.0])),
        freq=new_freq,
        wavelength=new_wavelength,
        is_frequency_scaled=wavelength_scaled,
    )


@add_method(IntensitySpectrum, "get_transform_limited_pulse")
def get_transform_limited_pulse(self, gate_level: Optional[float] = None):
    """Returns the transform-limited pulse corresponding to the spectrum.

    Args:
        self: the class
        gate_level (float): Apply a gate such that only values above gate_level*max(spectrum) are included

    Returns:
        ComplexEnvelope: the transform-limited pulse

    """
    if self.spectrum is not None:
        from ..spectrum import transform_limited_pulse_from_spectrometer

        spec = np.array(self.spectrum)
        lam = None
        if self.wavelength is None and self.freq is not None:
            lam = constants.speed_of_light / self.freq[self.freq > 0.0]
            spec = spec[self.freq > 0.0]
        elif self.wavelength is not None:
            lam = self.wavelength
        if lam is not None:
            if self.is_frequency_scaled:
                spec /= lam**2
            t, intensity = transform_limited_pulse_from_spectrometer(
                1e9 * lam, spec, gate_level
            )
            return ComplexEnvelope(envelope=np.sqrt(intensity), time=t, dt=t[1] - t[0])
    raise Exception("Missing data")


@add_method(ComplexEnvelope, "to_complex_spectrum")
def complex_envelope_to_complex_spectrum(
    self, padding_factor: int = 1
) -> ComplexSpectrum:
    """Returns a ComplexSpectrum based on the data."""
    return ComplexSpectrum(
        spectrum=np.fft.rfft(self.envelope, self.envelope.shape[0] * padding_factor),
        freq=np.fft.rfftfreq(self.envelope.shape[0] * padding_factor, self.dt)
        + self.carrier_frequency,
    )


@add_method(ComplexEnvelope, "to_waveform")
def complex_envelope_to_waveform(
    self, interpolation_factor: int = 1, CEP_shift: float = 0.0
) -> Waveform:
    """Returns a Waveform based on the data."""
    output_dt = self.dt / interpolation_factor
    output_time = self.time[0] + output_dt * np.array(
        range(self.time.shape[0] * interpolation_factor)
    )
    if interpolation_factor != 1:
        output_envelope = np.array(
            sig.resample(np.real(self.envelope), output_time.shape[0])
        ) + 1j * np.array(sig.resample(np.imag(self.envelope), output_time.shape[0]))
    else:
        output_envelope = self.envelope
    return Waveform(
        wave=np.real(
            np.exp(
                1j * self.carrier_frequency * 2 * np.pi * output_time - 1j * CEP_shift
            )
            * output_envelope
        ),
        time=output_time,
        dt=output_dt,
        is_uniformly_spaced=True,
    )



"""Functions for accessing and saving data."""

import io
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
import scipy.io as sio
from scipy import constants

from .. import spectrum
from ..numeric import interpolate
from .frog_data import FrogData, Spectrogram
from .interop import ComplexSpectrum, IntensitySpectrum, Waveform


def read_dwc(file_or_path, is_buffer: bool = False):
    """Reads files in the .dwc format produced by many FROG scanners.

    Args:
        file_or_path: path to the .dwc file
        is_buffer: set to true if you are giving a file in a bytes buffer instead of a path

    Returns:
        Spectrogram: the loaded data

    """
    if is_buffer:
        file_or_path = io.BytesIO(file_or_path)
        lines = file_or_path.readlines()
        delay_increment = float(lines[2].decode().strip().split("=")[1])
        file_or_path.seek(0)
    else:
        with open(file_or_path, "r") as file:
            lines = file.readlines()
            delay_increment = float(lines[2].strip().split("=")[1])

    wavelength_vector = pd.read_csv(
        file_or_path, skiprows=5, nrows=1, delimiter="\t", header=None
    ).values[0]

    if is_buffer:
        file_or_path.seek(0)

    data_array = pd.read_csv(
        file_or_path, skiprows=8, delimiter="\t", header=None, dtype=float
    ).values

    freqs, first_spec = spectrum.wavelength_to_frequency(
        wavelength_vector, data_array[:, 0]
    )
    if freqs is not None:
        data_array_freq = np.zeros((len(freqs), data_array.shape[1]))
        data_array_freq[:, 0] = first_spec
        for _i in range(data_array.shape[1]):
            _f, _dat = spectrum.wavelength_to_frequency(
                wavelength_vector, data_array[:, _i], freqs
            )
            data_array_freq[:, _i] = _dat
        dt = 1e-15 * delay_increment
        delays = dt * np.array(range(data_array.shape[1]))
        delays -= np.mean(delays)
        return Spectrogram(data=data_array_freq, time=delays, freq=freqs)
    raise Exception("Interpolation failure reading dwc file")


def load_mean_spectrum_from_scarab(
    filename_or_data_string: str, is_data_string: bool = False, header_size: int = 0
):
    """Load data produced by Scarab (Nick's C++ interface for Ocean Optics spectrometers).

    Args:
        filename_or_data_string (str): path to the file, or a bytestream containing the file
        is_data_string (bool): set to true if using a bytestream
        header_size (int): lines to skip before getting to the real data

    """
    if is_data_string:
        pd_input = pd.read_csv(
            io.StringIO(filename_or_data_string), sep="\t", skiprows=header_size
        )
        data = np.array(pd_input.iloc[:, :])
    else:
        data = np.loadtxt(filename_or_data_string, skiprows=header_size)

    return IntensitySpectrum(
        spectrum=np.mean(data[:, 1::], axis=1),
        wavelength=1e-9 * data[:, 0],
        freq=1e9 * constants.speed_of_light / data[:, 0],
        is_frequency_scaled=False,
        phase=None,
    )


def load_spectrum_from_text(
    filename_or_data_string: str,
    wavelength_multiplier: float = 1e-9,
    wavelength_field: str = "wavelength (nm)",
    spectrum_field: str = "intensity (a.u.)",
    sep: str = "\t",
    is_data_string: bool = False,
):
    """Load a spectrum contained in a text file.

    Args:
        filename_or_data_string (str): path to the file or a string containing the file
        wavelength_multiplier (float): multiplier to convert wavelength to m
        wavelength_field (str): name of the field in the data corresponding to wavelength
        spectrum_field (str): name of the field in the data corresponding to spectral intensity
        sep (str): column separator
        is_data_string (bool): set to True if the first input is the full file, not the path
    Returns:
        IntensitySpectrum: the intensity spectrum

    """
    if is_data_string:
        data = pd.read_csv(io.StringIO(filename_or_data_string), sep=sep)
    else:
        data = pd.read_csv(filename_or_data_string, sep=sep)
    wavelength = wavelength_multiplier * np.array(data[wavelength_field])
    freq = constants.speed_of_light / wavelength
    spectrum = np.array(data[spectrum_field])
    return IntensitySpectrum(
        spectrum=spectrum,
        wavelength=wavelength,
        freq=freq,
        phase=np.zeros(spectrum.shape, dtype=float),
    )


def load_waveform_from_text(
    filename: str,
    time_multiplier: float = 1e-15,
    time_field: str = "delay (fs)",
    wave_field: str = "field (a.u.)",
    sep="\t",
) -> Waveform:
    """Loads a waveform from a text file.

    Args:
        filename (str): path to the file
        time_multiplier (float): multiplier needed to convert the time unit in the file to seconds
        time_field (str): name (header) of the column containing the times
        wave_field (str): name (header) of the column containing the waveform
        sep (str): separator used in the file format (tab is default)

    Returns:
        Waveform: the waveform

    """
    data = pd.read_csv(filename, sep=sep)
    time = time_multiplier * data[time_field].to_numpy()
    wave = data[wave_field].to_numpy()
    dt = time[1] - time[0]
    diff_time = np.diff(time)
    uniform = bool(np.all(np.isclose(diff_time, diff_time[0])))
    return Waveform(wave=wave, time=time, dt=dt, is_uniformly_spaced=uniform)


def load_waves_from_matfile(filename: str, phase: Optional[float] = None):
    """Load the contents of an attolab scanner file in .mat format.

    Args:
        phase (float): phase to use when interpreting the lock-in data
        filename (str): path to the mat file
    Returns:
        time_delay: array of time delay values
        signal: signals corresponding to the time delays

    """
    datablob = sio.loadmat(filename)
    stage_position = datablob["xdata"][0, :]
    time_delay = -2e-3 * stage_position / 2.9979e8
    lia_x = datablob["x0"]
    lia_y = datablob["y0"]
    if phase is None:
        optimized_phase = np.atan2(np.sum(lia_y[:] ** 2), np.sum(lia_x[:] ** 2))
        signal = np.fliplr(
            lia_x * np.cos(optimized_phase) + lia_y * np.sin(optimized_phase)
        )
        return time_delay, signal
    signal = np.fliplr(lia_x * np.cos(phase) - lia_y * np.sin(phase))
    return time_delay, signal


def read_Trebino_FROG_matrix(filename: Path | str) -> Spectrogram:
    """Read a spectrogram file made by the Trebino FROG code.

    Args:
        filename (Path | str): the name (path) of the file

    """
    with open(filename, "r") as f:
        line = str(f.readline())
        line = line.split()
        n1 = int(line[0])
        n2 = int(line[1])
        line = str(f.readline())
        line = line.split()
    measured_data = pd.read_csv(filename, sep="\t", header=None, skiprows=2)
    measure = []
    raw_freq = (
        1e9 * constants.speed_of_light / np.array(measured_data[0][0:n2]).squeeze()
    )
    df = np.mean(np.diff(raw_freq))
    freq = raw_freq[0] + df * np.array(range(raw_freq.shape[0]))
    time = 1e-15 * np.array(measured_data[0][n2 : (n2 + n1)]).squeeze()
    for i in range(n1):
        measure.append(measured_data[0][(i + 2) * n2 : (i + 3) * n2])
    data = np.array(measure)
    return Spectrogram(data=data, time=time, freq=freq)


def read_Trebino_FROG_speck(filename: Path | str) -> ComplexSpectrum:
    """Read a .Speck file made by the Trebino FROG code.

    Args:
        filename (Path | str): the name (path) of the file

    """
    data = np.array(pd.read_csv(filename, sep="\t", header=None), dtype=float)
    raw_freq = 1e9 * constants.speed_of_light / data[:, 0]
    df = np.mean(np.diff(raw_freq))
    freq = np.linspace(0.0, raw_freq[-1], int(np.ceil(raw_freq[-1] / df)))
    spectrum = interpolate(freq, raw_freq, data[:, 3]) + 1j * interpolate(
        freq, raw_freq, data[:, 4]
    )
    return ComplexSpectrum(spectrum=spectrum, freq=freq)


def read_Trebino_FROG_data(filename: str) -> FrogData:
    """Read a set of data produced by the Trebino FROG reconstruction code.

    Args:
        filename: Base filename of the .bin file; e.g. if the data are mydata.bin.Speck.dat etc., this will be "mydata.bin"

    """
    spectrum = read_Trebino_FROG_speck(filename + ".Speck.dat")
    pulse = spectrum.to_centered_waveform()
    measured_spectrogram = read_Trebino_FROG_matrix(filename + ".A.dat")
    reconstructed_spectrogram = read_Trebino_FROG_matrix(filename + ".Arecon.dat")
    raw_speck = np.array(
        pd.read_csv(filename + ".Speck.dat", sep="\t", header=None), dtype=float
    )
    raw_ek = np.array(
        pd.read_csv(filename + ".Ek.dat", sep="\t", header=None), dtype=float
    )
    f0 = 1e9 * np.mean(constants.speed_of_light / raw_speck[:, 0])
    dt = 1e-15 * (raw_ek[1, 0] - raw_ek[0, 0])
    raw_reconstruction = raw_ek[:, 3] + 1.0j * raw_ek[:, 4]
    return FrogData(
        spectrum=spectrum,
        pulse=pulse,
        measured_spectrogram=measured_spectrogram,
        reconstructed_spectrogram=reconstructed_spectrogram,
        raw_reconstruction=raw_reconstruction,
        dt=dt,
        f0=float(f0),
    )



"""Waveform processing tools."""

import numpy as np
import scipy.optimize as opt
import scipy.signal as sig


def align_waves(
    waves, dt: float, frequency_roi_start: float, frequency_roi_stop: float
):
    """Align a set of waveforms, inside of a 2D numpy array.

    Args:
        waves: set of waveforms
        dt (float): time step (assuming constant spacing), (s)
        frequency_roi_start (float): start frequency of region of interest for alignment (Hz)
        frequency_roi_stop (float): stop frequency of region of interest for alignment (Hz)

    Returns:
        np.ndarray: set of aligned waves

    """
    waves_f = np.array(waves)

    # windowing and offset removal
    for i in range(waves_f.shape[0]):
        waves_f[i, :] = sig.windows.tukey(waves_f.shape[1]) * (
            waves_f[i, :] - np.mean(waves_f[i, :])
        )

    # find frequency roi
    f = np.fft.fftfreq(waves.shape[1], dt)
    f0 = np.argmin(np.abs(f - frequency_roi_start))
    f1 = np.argmin(np.abs(f - frequency_roi_stop))
    w_roi = 2 * np.pi * f[f0:f1]

    # get active spectral region
    waves_f = np.fft.fft(waves, axis=1)
    waves_roi = np.array(waves_f[:, f0:f1])
    waves_roi /= np.max(np.max(np.abs(waves_roi)))

    # apply tau phase shifts
    def apply_taus(spectrum, taus, w):
        spectrum_shifted = np.zeros(spectrum.shape, dtype=complex)
        for i in range(spectrum.shape[0]):
            spectrum_shifted[i, :] = np.exp(-1j * 1e-18 * taus[i] * w) * spectrum[i, :]
        return spectrum_shifted

    # return fitting weights
    def get_residual(taus):
        shifted = apply_taus(waves_roi, taus, w_roi)
        mean_amplitudes = np.mean(shifted, axis=0)
        return 5.0 - np.abs(mean_amplitudes)

    # apply fitting
    res = opt.least_squares(
        get_residual, np.zeros(waves.shape[0]), ftol=1e-12, max_nfev=16384
    )

    # remove mean shift
    res.x -= np.mean(res.x)

    print(f"Rms shift in attoseconds: {np.std(res.x)}")
    return np.real(np.fft.ifft(apply_taus(waves_f, res.x, 2 * np.pi * f)))



"""Contains functions for plotting, and for formatting plots, with the goal of more easily obtaining a consistent style across plots made by different co-authors."""

from .plot import Char, label_letter, set_style, showmo

__all__ = ["showmo", "set_style", "label_letter", "Char"]



import numpy as np
import pandas as pd
import scipy.constants as constants
import scipy.optimize as opt
import scipy.signal
from matplotlib import pyplot as plt

from ...numeric import fwhm


def get_fwhm(t, x, no_envelope: bool = False):
    """Calculates the full-width-at-half-maximum:
    FWHM is the full-width of the intensity profile I(t) = |A(t)|^2 at half-maximum, where A(t) is the envelope of the signal.

    Args:
        x: signal in time domain vector
        t: time vector

    Returns:
        fwhm: fwhm_val[0]

    if no_envelope == True it computes the FWHM without taking the square modulus and without computing envelope (assuming input is intensity envelope already)

    """
    if no_envelope and np.min(x) < 0:
        raise ValueError("tried to compute FWHM of an envelope with negative values")
    envelope = np.abs(scipy.signal.hilbert(x)) ** 2
    return fwhm(envelope, t[1] - t[0])


def gaussian(height, center_x, center_y, width_x, width_y, offset):
    """Returns a gaussian function with the given parameters."""
    width_x = float(width_x)
    width_y = float(width_y)
    return lambda x, y: offset + height * np.exp(
        -(((center_x - x) / width_x) ** 2 + ((center_y - y) / width_y) ** 2) / 2
    )


def eliminate_broken_pixels(data):
    """Eliminates the broken pixels of the camera rayCi I was using."""
    data[395:409, 950:1060] = 0
    return data


def moments(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution by calculating its
    moments.
    """
    total = data.sum()
    X, Y = np.indices(data.shape)
    x = (X * data).sum() / total
    y = (Y * data).sum() / total
    col = data[:, int(y)]
    width_x = np.sqrt(np.abs((np.arange(col.size) - x) ** 2 * col).sum() / col.sum())
    row = data[int(x), :]
    width_y = np.sqrt(np.abs((np.arange(row.size) - y) ** 2 * row).sum() / row.sum())
    height = data.max()
    return height, x, y, width_x, width_y, np.min(data)


def moments_peak(data):
    """Returns (height, x, y, width_x, width_y, np.min(data)), that is,
    the gaussian parameters of a 2D distribution by calculating its
    peak and moments.
    """
    X, Y = np.indices(data.shape)
    maxindex = np.unravel_index(np.argmax(data), data.shape)
    x, y = maxindex[0], maxindex[1]
    col = data[:, int(y)]
    width_x = np.sqrt(np.abs((np.arange(col.size) - x) ** 2 * col).sum() / col.sum())
    row = data[int(x), :]
    width_y = np.sqrt(np.abs((np.arange(row.size) - y) ** 2 * row).sum() / row.sum())
    height = data.max()
    return height, x, y, width_x, width_y, np.min(data)


def fitgaussian(data):
    """Returns (height, x, y, width_x, width_y)
    the gaussian parameters of a 2D distribution found by a fit.
    """
    params = moments_peak(data)

    def errorfunction(p):
        return np.ravel(gaussian(*p)(*np.indices(data.shape)) - data)

    p, success = opt.leastsq(errorfunction, params)
    return p


def plot_crosssect(data, fitfunct=None):
    """Plots the x- and y- cross-sections of a 2D array (image). The two cross-sections contain the maximum-valued pixel.

    Args:
        data: 2D array
        fitfunct: function (callable) object describing the image profile. Defaults to None

    """
    maxindex = np.unravel_index(np.argmax(data), data.shape)
    y, x = maxindex[0], maxindex[1]
    Y, X = np.indices(data.shape)

    fig, ax = plt.subplots(1, 2)
    ax[0].plot(X[y], data[y])
    if fitfunct is not None:
        ax[0].plot(X[y], fitfunct(X[y] * 0.0 + y, X[y]))
    ax[1].plot(Y[:, x], data[:, x])
    if fitfunct is not None:
        ax[1].plot(Y[:, x], fitfunct(Y[:, x], Y[:, x] * 0.0 + x))
    return fig


def cut_tail(data, radius, pixelsize=1.0, center=None):
    radius = radius / pixelsize
    if center is None:
        center = np.unravel_index(np.argmax(data), data.shape)
    y, x = center[0], center[1]
    Y, X = np.indices(data.shape)
    cut_data = data.copy()
    cut_data[np.sqrt((Y - y) ** 2 + (X - x) ** 2) > radius] = 0.0
    return cut_data


def cut_trace(t, x, radius, center=None):
    if center is None:
        center = np.argmax(abs(x))
    cut_trace = x.copy()
    cut_trace[abs(t - t[center]) > radius] = 0
    return cut_trace


def integral2d(data, pixelsize=1.0):
    return data.sum() * pixelsize**2


def trace_integral(t, x):
    dt = np.gradient(t)
    f = x  # np.array([(i+j)/2 for (i,j) in zip(x[:-1], x[1:])])
    return (dt * f).sum()


def twoD_Gaussian(xy, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    x, y = xy
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta) ** 2) / (2 * sigma_x**2) + (np.sin(theta) ** 2) / (
        2 * sigma_y**2
    )
    b = -(np.sin(2 * theta)) / (4 * sigma_x**2) + (np.sin(2 * theta)) / (4 * sigma_y**2)
    c = (np.sin(theta) ** 2) / (2 * sigma_x**2) + (np.cos(theta) ** 2) / (
        2 * sigma_y**2
    )
    g = offset + amplitude * np.exp(
        -(a * ((x - xo) ** 2) + 2 * b * (x - xo) * (y - yo) + c * ((y - yo) ** 2))
    )
    return g.ravel()


def profile_analysis(
    profile_file,
    trace_file=None,
    magnification=1.0,
    pixelsize=5.0,
    power=20.0,
    reprate=4.0,
    ROI_diam: int = 100,
    forced_background=None,
    trace_cutoff_radius=50.0,
    cutoff_gaus_fit_profile=25.0,
):
    """If trace_file is None, this function only loads, fits and plots the beam profile; if trace_file is given, the function additionally calculates the peak intensity and field.

    Args:
        profile_file: name of the file containing the beam profile; the file (txt or csv) contains the 2D array of the beam profile
        trace_file: None; name of the file containing the trace; the file (txt or csv) contains the time and field arrays as tab-separated columns
        magnification: 1.; ratio of lens-camera/lens-z0 lengths; This is used to give the correct size of the beam profile at z0 in um
        pixelsize: 5. um; pixel size of the camera
        power: 20. mW; power of the laser beam; this is used to calculate the peak intensity
        reprate: 4 kHz; repetition rate of the laser; this is used to calculate the peak intensity
        ROI_diam: 100; diameter of the region of interest in pixels; the center of the ROI is the maximum of the camera data (which is assumed to be not saturated)
        cutoff_gaus_fit_profile: 25. pixels; the 2d gaussian is fitted to the beam profile up to this distance from the peak
        forced_background: None; if not None, the subtracted background is set to this value; otherwise it's computed from the outer ring of the ROI (camera data)
        trace_cutoff_radius: 50. fs; for intensity calculation, the integral of the trace is computed only in the range t[peak]-trace_cutoff_radius < t < t[peak]+trace_cutoff_radius

    """
    pixelsize = pixelsize / magnification
    data = pd.read_csv(profile_file)

    dataval = np.array(data.values)
    # print('eliminating fixed broken pixels rayCi camera')
    # dataval=eliminate_broken_pixels(dataval)
    maxindex = np.unravel_index(np.argmax(dataval), dataval.shape)
    print("diameter of Region Of Interest (ROI): ", ROI_diam, " pixels")
    dataval = np.array(
        data.values[
            maxindex[0] - int(ROI_diam / 2) : maxindex[0] + int(ROI_diam / 2),
            maxindex[1] - int(ROI_diam / 2) : maxindex[1] + int(ROI_diam / 2),
        ]
    )
    maxindex = np.unravel_index(np.argmax(dataval), dataval.shape)

    background = 0.0
    count = 0
    print(
        "calculating background intensity outside a circle whose diameter is 3/4 of the ROI diameter"
    )
    for i, j in np.ndindex(dataval.shape):
        if (i - maxindex[0]) ** 2 + (j - maxindex[1]) ** 2 > (ROI_diam * 3 / 8) ** 2:
            background += dataval[i, j]
            count += 1
    background /= count

    if forced_background is not None:
        background = forced_background
        print("subtracting background intensity: ", background)
    else:
        print("subtracting background intensity: ", background)

    dataval = np.maximum(dataval - background, dataval * 0.0)

    # gaussian fit
    param = moments_peak(dataval)
    param = fitgaussian(cut_tail(dataval, cutoff_gaus_fit_profile))
    X, Y = np.indices(dataval.shape)
    fitted_funct = gaussian(*param)(X, Y)
    print(
        "beam w (um), resp. in x and y axis:",
        2 * param[3] * pixelsize,
        2 * param[4] * pixelsize,
    )
    print(
        "beam 1/e^2 diameter [um]: ", 4 * param[3] * pixelsize, 4 * param[4] * pixelsize
    )

    if trace_file is not None:
        trace = pd.read_csv(trace_file, sep="\t")

        # time in fs, field in a.u.
        intensity = (
            power
            * 1e-3
            / reprate
            * 1e-3
            / integral2d(fitted_funct, pixelsize=pixelsize)
            * np.max(fitted_funct)
            * 1e8
            / trace_integral(
                trace["delay (fs)"],
                cut_trace(
                    trace["delay (fs)"], trace["field (a.u.)"] ** 2, trace_cutoff_radius
                ),
            )
            * np.max(trace["field (a.u.)"] ** 2)
            * 1e15
        )

        intensityFromRawCameraData = (
            power
            * 1e-3
            / reprate
            * 1e-3
            / integral2d(
                dataval[
                    maxindex[0] - int(ROI_diam / 2) : maxindex[0] + int(ROI_diam / 2),
                    maxindex[1] - int(ROI_diam / 2) : maxindex[1] + int(ROI_diam / 2),
                ],
                pixelsize=pixelsize,
            )
            * np.max(dataval)
            * 1e8
            / trace_integral(
                trace["delay (fs)"],
                cut_trace(
                    trace["delay (fs)"], trace["field (a.u.)"] ** 2, trace_cutoff_radius
                ),
            )
            * np.max(trace["field (a.u.)"] ** 2)
            * 1e15
        )

        temporal_fwhm = get_fwhm(
            np.array(trace["delay (fs)"]), np.array(trace["field (a.u.)"])
        )
        intensityFromTemporalFWHM = (
            power
            * 1e-3
            / reprate
            * 1e-3
            / integral2d(
                dataval[
                    maxindex[0] - int(ROI_diam / 2) : maxindex[0] + int(ROI_diam / 2),
                    maxindex[1] - int(ROI_diam / 2) : maxindex[1] + int(ROI_diam / 2),
                ],
                pixelsize=pixelsize,
            )
            * np.max(dataval)
            * 1e8
            / (np.sqrt(2 * np.pi) * temporal_fwhm / np.sqrt(8 * np.log(2)))
            * 1e15
        )

        print(
            "\nusing the gaussian fit of the beam profile and the integral of the trace^2:\n "
            "peak intensity (sub-cycle, W/cm2): {i:.3e}".format(i=intensity)
        )
        peak_field = np.sqrt(
            intensity * 1e4 / constants.speed_of_light / constants.epsilon_0
        )
        print("peak field (sub-cycle, V/m): {f:.3e}".format(f=peak_field))

        print(
            "\nusing directly camera data, and the integral of the trace^2:\n "
            "peak intensity (sub-cycle, W/cm2): {i:.3e}".format(
                i=intensityFromRawCameraData
            )
        )
        peak_field = np.sqrt(
            intensityFromRawCameraData
            * 1e4
            / constants.speed_of_light
            / constants.epsilon_0
        )
        print("peak field (sub-cycle, V/m): {f:.3e}".format(f=peak_field))

        print(
            "\nusing directly camera data, and the temporal FWHM retrieved from the trace:\n "
            "peak intensity (sub-cycle, W/cm2): {i:.3e}".format(
                i=intensityFromTemporalFWHM
            )
        )
        peak_field = np.sqrt(
            intensityFromTemporalFWHM
            * 1e4
            / constants.speed_of_light
            / constants.epsilon_0
        )
        print("peak field (sub-cycle, V/m): {f:.3e}".format(f=peak_field))

        fig, ax = plt.subplots(1, 1)
        time = trace["delay (fs)"]
        field = trace["field (a.u.)"]
        i_peak = np.argmax(field**2)
        ax.plot(
            time[np.abs(time - time[i_peak]) < trace_cutoff_radius],
            field[np.abs(time - time[i_peak]) < trace_cutoff_radius] ** 2,
        )
        plt.show()

    fig, ax = plt.subplots(1, 1)
    ax.imshow(dataval)
    plt.show()

    plot_crosssect(dataval, fitfunct=gaussian(*param))



"""Define the base version of the Waveform class, which is extended in interop.py."""

import copy
from dataclasses import dataclass

import numpy as np
import scipy.signal as sig

from ..numeric import (
    find_maximum_location,
    fwhm,
    interpolate,
)
from .decorators import yaml_io


@yaml_io
@dataclass(slots=True)
class Waveform:
    """Contains data describing an electric field waveform.
    In SI units.

    Attributes:
        wave (np.ndarray): the waveform data
        time (np.ndarray): time vector corresponding to the wave array
        dt (float): the time step of time, if it is uniformly spaced
        is_uniformly_spaced (bool): says whether or not the time steps are uniform

    """

    wave: np.ndarray
    time: np.ndarray
    dt: float
    is_uniformly_spaced: bool = False

    def lock(self):
        """Make the data immutable."""
        self.wave.setflags(write=False)
        self.time.setflags(write=False)

    def copy(self):
        """Returns a deep copy."""
        return copy.deepcopy(self)

    def time_fs(self):
        """Returns the time in femtoseconds."""
        if self.time is not None:
            return 1e15 * self.time
        raise Exception("No data")

    def to_uniformly_spaced(self):
        """Returns a version of the struct with uniformly spaced time."""
        if self.is_uniformly_spaced:
            return self
        timesteps = np.abs(np.diff(self.time))
        new_dt = np.min(timesteps[timesteps > 0.0])
        new_time_length = self.time[-1] - self.time[0]
        new_time = self.time[0] + new_dt * np.array(
            range(int(new_time_length / new_dt))
        )
        return Waveform(
            wave=interpolate(new_time, self.time, self.wave),
            time=new_time,
            dt=new_dt,
            is_uniformly_spaced=True,
        )

    def to_windowed(self, window_desc):
        """Create a windowed version of the waveform. Output will be uniformly spaced in time, even if current state isn't.

        Args:
            window_desc: String or tuple describing the desired window in the same format as scipy.signal.windows.get_window.

        Examples:
            >>> waveform_with_tukey_window = waveform.to_windowed('tukey')
            >>> waveform_with_supergaussian_window = waveform.to_windowed(('general_gaussian', 2, 500))

        """
        uniform_self = self.to_uniformly_spaced()
        new_wave = uniform_self.wave * sig.windows.get_window(
            window_desc, Nx=uniform_self.wave.shape[0]
        )
        return Waveform(
            wave=new_wave,
            time=uniform_self.time,
            dt=uniform_self.dt,
            is_uniformly_spaced=True,
        )

    def to_bandpassed(self, frequency: float, sigma: float, order: int = 4):
        r"""Apply a bandpass filter, with the same spec as to_bandpassed method of ComplexSpectrum:
        $e^{\frac{(f-f_0)^r}{2\sigma^r}}$
        where $f_0$ is the frequency argument, $\sigma$ is the sigma argument, and r is the order argument.

        Args:
            frequency: the central frequency (Hz) of the bandpass
            sigma: the width of the bandpass (Hz)
            order: the order of the supergaussian

        """
        return (
            self.to_complex_spectrum()
            .to_bandpassed(frequency, sigma, order)
            .to_waveform()
        )

    def to_normalized(self):
        """Return a normalized version of the waveform."""
        max_loc, max_val = find_maximum_location(
            np.abs(np.array(sig.hilbert(self.wave)))
        )
        return Waveform(
            wave=self.wave / max_val,
            time=np.array(self.time),
            dt=self.dt,
            is_uniformly_spaced=self.is_uniformly_spaced,
        )

    def get_envelope_fwhm(self) -> float:
        """Get the full-width-at-half-maximum of the intensity envelope.

        Returns:
            float: the FWHM

        """
        return self.to_complex_envelope().get_fwhm()

    def get_field_squared_fwhm(self):
        """Get the full-width-at-half-maximum of the square of the field.

        Returns:
            float: the FWHM

        """
        uniform_self = self.to_uniformly_spaced()
        return fwhm(uniform_self.wave**2, uniform_self.dt)



"""Handle data sets related to producing a calibration, and complete the SpectrometerCalibration class."""

from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
from scipy import constants

from ..numeric import interpolate
from .decorators import add_method, yaml_io
from .frog_data import Spectrogram
from .interop import IntensitySpectrum
from .spectrometer_calibration import SpectrometerCalibration


@add_method(SpectrometerCalibration, "apply_to_spectrum")
def calibration_apply_to_spectrum(self, spectrum_in):
    """Apply the calibration to an intensity spectrum.

    Args:
        self: the calibration (implicit)
        spectrum_in (IntensitySpectrum): the spectrum to be calibrated

    Returns:
        IntensitySpectrum: the calibrated spectrum

    """
    intensity_out = spectrum_in.spectrum * self.intensity_factors
    return IntensitySpectrum(
        spectrum=intensity_out,
        phase=spectrum_in.phase,
        wavelength=self.corrected_wavelengths,
        freq=self.corrected_frequencies,
    )


@add_method(SpectrometerCalibration, "apply_to_spectrogram")
def calibration_apply_to_spectrogram(self, spectrogram_in):
    """Apply the calibration to an intensity spectrum.

    Args:
        self: the calibration (implicit)
        spectrogram_in (Spectrogram): the spectrogram to be calibrated

    Returns:
        Spectrogram: the calibrated spectrogram

    """
    original_freqs = constants.speed_of_light / self.original_wavelengths
    original_freq_projection = interpolate(
        spectrogram_in.freq,
        original_freqs,
        self.corrected_frequencies,
        inputs_are_sorted=False,
    )
    data_out = np.zeros(spectrogram_in.data.shape, dtype=float)
    intensity_factors = interpolate(
        spectrogram_in.freq,
        self.corrected_frequencies,
        self.intensity_factors,
        inputs_are_sorted=False,
    )
    for i in range(spectrogram_in.data.shape[1]):
        data_out[:, i] = intensity_factors * interpolate(
            spectrogram_in.freq,
            original_freq_projection,
            np.array(spectrogram_in.data[:, i]),
        )

    return Spectrogram(
        data=data_out, freq=spectrogram_in.freq, time=spectrogram_in.time
    )


@yaml_io
@dataclass
class CalibrationInput:
    """Input parameters for fitting a calibration curve from a data set."""

    wavelength_center: float
    wavelength_offset: float
    wavelength_slope: float
    amplitude_center: float
    amplitude_multiplier: float
    amplitude_slope: float
    amplitude_width: float
    amplitude_order: float
    roi_lowest: float
    roi_highest: float
    noise_level: float

    def get_wavelength_array(self):
        """Produce an array of the wavelength parameters for fitting routines."""
        return np.array(
            [self.wavelength_center, self.wavelength_offset, self.wavelength_slope]
        )

    def get_amplitude_array(self):
        """Produce an array of the amplitude parameters for the fitting routine."""
        return np.array(
            [
                self.amplitude_center,
                self.amplitude_multiplier,
                self.amplitude_slope,
                self.amplitude_width,
                self.amplitude_order,
            ]
        )

    def plot(
        self,
        measurement: IntensitySpectrum,
        reference: IntensitySpectrum,
        plot_xmin=None,
        plot_xmax=None,
    ):
        """Combare the reference and measurement."""
        fig, ax = plt.subplots(1, 2, figsize=(8, 4))
        ax[0].plot(
            measurement.wavelength_nm(), measurement.spectrum, label="Measurement"
        )
        ax[0].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
        ax[0].set_xlabel("Wavelength (nm)")
        ax[0].set_ylabel("Intensity (Arb. unit)")
        ax[0].legend()
        initial_guess_calibration = generate_calibration_from_coeffs(
            self.get_amplitude_array(),
            self.get_wavelength_array(),
            measurement.wavelength,
        )

        initial_guess_spectrum = initial_guess_calibration.apply_to_spectrum(
            measurement
        )
        ax[1].plot(
            initial_guess_spectrum.wavelength_nm(),
            initial_guess_spectrum.spectrum,
            label="Initial guess",
        )
        ax[1].plot(reference.wavelength_nm(), reference.spectrum, label="Reference")
        ax[1].set_ylabel("Intensity (Arb. unit)")
        ax[1].set_xlabel("Wavelength (nm)")
        ax[1].legend()

        if plot_xmin is not None and plot_xmax is not None:
            for a in ax:
                a.set_xlim(plot_xmin, plot_xmax)

        return fig


def generate_response_curve(
    wavelength: np.ndarray, coefficients: np.ndarray
) -> np.ndarray:
    """Turn the parameter efficients array and wavelengths into a response curve."""
    relative_wl = wavelength - coefficients[0]
    taylor_series = coefficients[1] + 0.5 * relative_wl * coefficients[2]
    gaussian = np.exp(
        -(np.abs(relative_wl) ** coefficients[4])
        / (2 * coefficients[3] ** coefficients[4])
    )
    return taylor_series * gaussian


def get_new_wavelength(wavelengths_micron, taylor_coefficients_micron):
    """Apply the wavelength adjustment coefficients to the input wavelengths."""
    distance = wavelengths_micron - taylor_coefficients_micron[0]
    taylor_shift = taylor_coefficients_micron[1] * np.ones(distance.shape, dtype=float)
    for i in range(2, len(taylor_coefficients_micron)):
        taylor_shift += distance ** (i - 1) * taylor_coefficients_micron[i]

    new_wavelengths = 1e-6 * (wavelengths_micron + taylor_shift)
    new_freq = constants.speed_of_light / new_wavelengths
    return new_wavelengths, new_freq


def fit_calibration_amplitude_model(
    measurement, reference, wavelength_coeffs, amplitude_guess, roi
):
    """Fit the model calibration."""
    initial_cal = generate_calibration_from_coeffs(
        amplitude_guess, wavelength_coeffs, measurement.wavelength
    )
    reference_shift = reference.to_interpolated_wavelength(
        initial_cal.corrected_wavelengths
    )

    def residual_amp(coeff):
        cal = generate_calibration_from_coeffs(
            coeff, wavelength_coeffs, measurement.wavelength
        )
        calibrated = cal.apply_to_spectrum(measurement)
        residuals = calibrated.spectrum - reference_shift.spectrum
        return residuals[
            (calibrated.wavelength > roi[0]) & (calibrated.wavelength < roi[1])
        ]

    amplitude_result = scipy.optimize.least_squares(
        residual_amp, amplitude_guess, ftol=1e-12, max_nfev=16384
    )

    return amplitude_result.x


def generate_calibration_from_coeffs(amplitude_coeffs, wavelength_coeffs, wavelengths):
    """Generate a model calibration."""
    new_wavelengths, new_freqs = get_new_wavelength(
        wavelengths * 1e6, wavelength_coeffs
    )
    intensity_factors = 1.0 / generate_response_curve(
        wavelength=new_wavelengths * 1e6, coefficients=amplitude_coeffs
    )
    return SpectrometerCalibration(
        intensity_factors=intensity_factors,
        original_wavelengths=wavelengths,
        corrected_wavelengths=new_wavelengths,
        corrected_frequencies=new_freqs,
    )


@yaml_io
@dataclass
class CalibrationDataset:
    """Collects all of the data related to a spectrometer calibration."""

    measurement: IntensitySpectrum
    reference: IntensitySpectrum
    input_parameters: CalibrationInput
    parameterized_calibration: SpectrometerCalibration
    measurement_with_parameterized_calibration: IntensitySpectrum
    residuals: np.ndarray
    final_calibration: SpectrometerCalibration

    def plot(self, plot_xmin=None, plot_xmax=None):
        """Plot the calibration."""
        fig, ax = plt.subplots(2, 3, figsize=(16, 9))
        ax[0, 0].plot(
            self.measurement.wavelength_nm(),
            self.measurement.spectrum,
            label="Measurement",
        )
        ax[0, 0].plot(
            self.reference.wavelength_nm(), self.reference.spectrum, label="Reference"
        )
        ax[0, 0].set_xlabel("Wavelength (nm)")
        ax[0, 0].set_ylabel("Intensity (Arb. unit)")
        ax[0, 0].legend()
        initial_guess_calibration = generate_calibration_from_coeffs(
            self.input_parameters.get_amplitude_array(),
            self.input_parameters.get_wavelength_array(),
            self.measurement.wavelength,
        )
        initial_guess_spectrum = initial_guess_calibration.apply_to_spectrum(
            self.measurement
        )
        ax[0, 1].plot(
            initial_guess_spectrum.wavelength_nm(),
            initial_guess_spectrum.spectrum,
            label="Initial guess",
        )
        ax[0, 1].plot(
            self.reference.wavelength_nm(), self.reference.spectrum, label="Reference"
        )
        ax[0, 1].set_ylabel("Intensity (Arb. unit)")
        ax[0, 1].set_xlabel("Wavelength (nm)")
        ax[0, 1].legend()

        ax[0, 2].plot(
            self.measurement_with_parameterized_calibration.wavelength_nm(),
            self.measurement_with_parameterized_calibration.spectrum,
            label="Model calibrated",
        )
        ax[0, 2].plot(
            self.reference.wavelength_nm(), self.reference.spectrum, label="Reference"
        )
        ax[0, 2].set_ylabel("Intensity (Arb. unit)")
        ax[0, 2].set_xlabel("Wavelength (nm)")
        ax[0, 2].legend()

        ax[1, 0].plot(
            self.measurement_with_parameterized_calibration.wavelength_nm(),
            self.residuals,
            label="Residual",
        )
        ax[1, 0].set_ylim(-0.1, 0.1)
        ax[1, 0].set_ylabel("Intensity (Arb. unit)")
        ax[1, 0].set_xlabel("Wavelength (nm)")
        ax[1, 0].legend()

        second_calibration = self.final_calibration.apply_to_spectrum(self.measurement)
        ax[1, 1].plot(
            second_calibration.wavelength_nm(),
            second_calibration.spectrum,
            label="Calibrated",
        )
        ax[1, 1].plot(
            self.reference.wavelength_nm(), self.reference.spectrum, label="Reference"
        )
        ax[1, 1].set_ylabel("Intensity (Arb. unit)")
        ax[1, 1].set_xlabel("Wavelength (nm)")
        ax[1, 1].legend()
        ax[1, 2].plot(
            self.final_calibration.corrected_wavelengths * 1e9,
            self.final_calibration.intensity_factors,
            label="Final weights",
        )
        ax[1, 2].set_ylabel("Final weights")
        ax[1, 2].set_xlabel("Wavelength (nm)")
        ax[1, 2].legend()

        if plot_xmin is not None and plot_xmax is not None:
            for a in ax:
                for b in a:
                    b.set_xlim(plot_xmin, plot_xmax)

        return fig

    @staticmethod
    def generate(
        measurement: IntensitySpectrum,
        reference: IntensitySpectrum,
        input_parameters: CalibrationInput,
    ):
        """Create a SpectrometerCalibration from a measured lamp spectrum and reference spectrom, and a set of parameters.

        Args:
            measurement (IntensitySpectrum): Lamp spectrum measured with the spectrometer in question
            reference (IntensitySpectrum): Known reference spectrum for the lamp
            input_parameters (CalibrationInput): Set of input parameters for the calibration fitting

        Returns:
            SpectrometerCalibration: The resulting calibration

        """
        new_guess = fit_calibration_amplitude_model(
            measurement=measurement,
            reference=reference,
            wavelength_coeffs=input_parameters.get_wavelength_array(),
            amplitude_guess=input_parameters.get_amplitude_array(),
            roi=np.array([input_parameters.roi_lowest, input_parameters.roi_highest]),
        )

        parameterized_calibration = generate_calibration_from_coeffs(
            new_guess, input_parameters.get_wavelength_array(), measurement.wavelength
        )
        measurement_with_parameterized_calibration = (
            parameterized_calibration.apply_to_spectrum(measurement)
        )

        ref_on_new_axis = reference.to_interpolated_wavelength(
            measurement_with_parameterized_calibration.wavelength
        )
        measurement_shifted = measurement.to_corrected_wavelength(
            measurement_with_parameterized_calibration.wavelength
        )
        residuals = (
            measurement_with_parameterized_calibration.spectrum
            - ref_on_new_axis.spectrum
        )

        new_weights = (
            parameterized_calibration.intensity_factors
            - measurement_shifted.spectrum
            * residuals
            / (measurement_shifted.spectrum**2 + input_parameters.noise_level)
        )

        final_calibration = SpectrometerCalibration(
            intensity_factors=new_weights,
            original_wavelengths=measurement.wavelength,
            corrected_frequencies=parameterized_calibration.corrected_frequencies,
            corrected_wavelengths=parameterized_calibration.corrected_wavelengths,
        )

        return CalibrationDataset(
            measurement=measurement,
            reference=reference,
            input_parameters=input_parameters,
            parameterized_calibration=parameterized_calibration,
            measurement_with_parameterized_calibration=measurement_with_parameterized_calibration,
            residuals=residuals,
            final_calibration=final_calibration,
        )



"""Tools for setting the style and appearance of plots."""

import io
from dataclasses import dataclass

import marimo as mo
import matplotlib.pyplot as plt
from cycler import cycler
from matplotlib.axes import Axes


def showmo():
    """Display a plot as an svg in marimo notebooks."""
    svg_buffer = io.StringIO()
    plt.savefig(svg_buffer, format="svg")
    return mo.output.append(mo.Html(svg_buffer.getvalue()))


def set_style(theme: str = "light", font_size: int = 11):
    """Set colors and fonts for matplotlib plots.

    Args:
        theme (str): Select color theme.
                    Options:
                        ```light```: color-blind friendly colors (default)
                        ```nick_dark```: dark mode that matches Nick's slides
        font_size (int): Font size to apply to all plots.

    """
    plt.rcParams.update(
        {
            "font.size": font_size,
            "xtick.labelsize": 0.9 * font_size,
            "ytick.labelsize": 0.9 * font_size,
            "legend.fontsize": 0.9 * font_size,
            "figure.autolayout": True,
        }
    )
    match theme:
        case "nick_dark":
            plt.rcParams.update(
                {
                    "font.sans-serif": [
                        "Helvetica",
                        "Nimbus Sans L",
                        "Arial",
                        "Verdana",
                        "Nimbus Sans L",
                        "DejaVu Sans",
                        "Liberation Sans",
                        "Bitstream Vera Sans",
                        "sans-serif",
                    ],
                    "font.family": "sans-serif",
                    "axes.prop_cycle": cycler(
                        color=["cyan", "magenta", "orange", "blueviolet", "lime"]
                    ),
                    "figure.facecolor": "#171717",
                    "figure.edgecolor": "#171717",
                    "savefig.facecolor": "#171717",
                    "savefig.edgecolor": "#171717",
                    "axes.facecolor": "black",
                    "text.color": "white",
                    "axes.edgecolor": "white",
                    "axes.labelcolor": "white",
                    "xtick.color": "white",
                    "ytick.color": "white",
                    "grid.color": "white",
                    "lines.color": "white",
                    "image.cmap": "magma",
                }
            )
        # Light case is combined with _, which will capture anything else that didn't match.
        # It must be the last case, for that reason.
        case "light" | _:
            plt.rcParams.update(
                {
                    "font.sans-serif": [
                        "Helvetica",
                        "Nimbus Sans L",
                        "Arial",
                        "Verdana",
                        "DejaVu Sans",
                        "Liberation Sans",
                        "Bitstream Vera Sans",
                        "sans-serif",
                    ],
                    "font.family": "sans-serif",
                    # colorblind-friendly color cycle from https://gist.github.com/thriveth/8560036
                    "axes.prop_cycle": cycler(
                        color=[
                            "#377eb8",
                            "#ff7f00",
                            "#4daf4a",
                            "#f781bf",
                            "#a65628",
                            "#984ea3",
                            "#999999",
                            "#e41a1c",
                            "#dede00",
                        ]
                    ),
                    "figure.facecolor": "white",
                    "figure.edgecolor": "white",
                    "savefig.facecolor": "white",
                    "savefig.edgecolor": "white",
                    "axes.facecolor": "white",
                    "text.color": "black",
                    "axes.edgecolor": "black",
                    "axes.labelcolor": "black",
                    "xtick.color": "black",
                    "ytick.color": "black",
                    "grid.color": "black",
                    "lines.color": "black",
                    "image.cmap": "viridis",
                }
            )


def label_letter(
    letter: str = "a",
    axis: Axes = plt.gca(),
    style: str = "Nature",
    x_position: float = -0.14,
    y_position: float = 1.08,
):
    """Put a letter in the corner of a set of axes to label them.

    Args:
        letter (str): The letter to use
        axis: The axes to use (default is current ones)
        style (str): The journal style to apply. Options are ```Nature```, ```Science```, and ```OSA```
        x_position (float): where to put the label horizontally relative to the axes of the figure
        y_position (float): vertical position

    """
    letter_string = f"{letter}"
    fontweight = "normal"
    match style:
        case "Nature":
            letter_string = letter_string.lower()
            fontweight = "bold"
        case "Science":
            letter_string = letter_string.upper()
            fontweight = "bold"
        case "OSA":
            letter_string = "(" + letter_string.lower() + ")"
            fontweight = "bold"

    axis.text(
        x_position,
        y_position,
        letter_string,
        ha="center",
        transform=axis.transAxes,
        fontweight=fontweight,
    )


@dataclass
class Char:
    """Contains useful characters and strings for plot labels.

    Examples:
        >>> plt.xlabel(aw.plot.Char.wavelength_micron)
        >>> plt.ylabel(f'{aw.plot.Char.theta} ({aw.plot.Char.degrees})')

    """

    mu: str = "\u03bc"
    phi: str = "\u03a6"
    lam: str = "\u03bb"
    theta: str = "\u03b8"
    micron: str = "\u03bcm"
    degrees: str = "\u00b0"
    angstrom: str = "\u212b"
    wavelength_micron: str = "Wavelength (\u03bcm)"
    energy_microjoule: str = "Energy (\u03bcJ)"



"""For processing Amelie's data."""

import numpy as np
import scipy.optimize as opt
import scipy.signal as sig


def mean_offset_tukey(signal):
    """Apply DC offset removal and a Tukey window to a waveform.

    Args:
        signal: the array describing the measured waveform

    Returns:
        np.ndarray: the modified signal

    """
    mean_signal = np.mean(signal, axis=0)
    mean_signal -= np.mean(mean_signal)
    mean_signal *= sig.windows.tukey(mean_signal.shape[0])
    return mean_signal


def filtered_impulse_response(
    E_signal,
    E_reference,
    dt: float,
    filter_f0: float,
    filter_sigma: float,
    filter_order: int,
):
    """Calculate the filtered impulse response given by a signal and reference pair, with applied bandpass filter.

    This function takes two waveforms, E_signal and E_reference, and a bandpass filter, and returns the
    impulse response function of whatever caused the waveform to change

    Args:
        E_signal: the signal waveform array
        E_reference: the reference waveform
        dt (float): the time step
        filter_f0 (float): central frequency of the bandpass
        filter_sigma (float): width of the bandpass
        filter_order (int): order of the bandpass

    Returns:
        np.ndarray: the impuse response

    """
    E_signal_f = np.fft.fft(E_signal)
    E_reference_f = np.fft.fft(E_reference)
    f = np.fft.fftfreq(E_signal.shape[0], dt)
    bandpass = np.exp(
        -((f - filter_f0) ** filter_order) / (2 * filter_sigma**filter_order)
    )
    return np.fft.ifft(bandpass * np.nan_to_num(E_signal_f / E_reference_f))


def shift_phase_amplitude(
    Et, dt: float, time_shift: float, phase_shift: float, amplitude_factor: float
):
    """Apply a time shift, phase shift, and amplitude factor to a field.

    Args:
        Et: the waveform to be modified
        dt: the time step of the waveform
        time_shift (float): the time delay to apply
        phase_shift (float): the phase shift to apply (radians)
        amplitude_factor (float): the amplitude factor to apply (unitless)

    Returns:
        np.ndarray: the modified waveform

    """
    Et_f = np.fft.fft(Et)
    f = np.fft.fftfreq(Et.shape[0], dt)
    return np.fft.ifft(
        amplitude_factor
        * np.exp(1j * (phase_shift - (time_shift * 2 * np.pi) * f))
        * Et_f
    )


def minimize_response_difference(response, reference):
    """Shift the delay, phase, and amplitude of the reference field to minimize the squared-intensity-envelope
    after subtracting them.

    Args:
        response: The signal waveform
        reference: The reference waveform

    Returns:
        np.ndarray: The modified reference field

    """
    # construct initial guess of the time, phase, and amplitude offsets
    peak_location_response = np.argmax(np.abs(response))
    peak_location_reference = np.argmax(np.abs(reference))
    start_phase_shift = np.angle(response[peak_location_response]) - np.angle(
        reference[peak_location_reference]
    )
    start_time_shift = peak_location_response - peak_location_reference
    start_amplitude = np.abs(reference[peak_location_reference]) / np.abs(
        response[peak_location_response]
    )
    start_values = [start_time_shift, start_phase_shift, start_amplitude]

    # function to get the envelope (to 4th power) of the difference between reference and response
    def get_residual(parameters):
        adjusted_reference = shift_phase_amplitude(
            reference, 1.0, parameters[0], parameters[1], parameters[2]
        )
        return np.abs(response - adjusted_reference) ** 4

    # minimize residuals
    res = opt.least_squares(get_residual, start_values, ftol=1e-12, max_nfev=16384)
    return shift_phase_amplitude(reference, 1.0, res.x[0], res.x[1], res.x[2])


def get_effective_response(
    E_signal,
    E_reference,
    dt: float,
    filter_f0: float,
    filter_sigma: float,
    filter_order: int,
):
    """Calculate the effective impulse response with the reactive response removed.

    Args:
        E_signal: the signal waveform array
        E_reference: the reference waveform
        dt (float): the time step
        filter_f0 (float): central frequency of the bandpass
        filter_sigma (float): width of the bandpass
        filter_order (int): order of the bandpass

    Returns:
        np.ndarray: the impuse response

    """
    analytic_signal = sig.hilbert(np.real(E_signal))
    analytic_reference = sig.hilbert(np.real(E_reference))

    response = filtered_impulse_response(
        analytic_signal, analytic_reference, dt, filter_f0, filter_sigma, filter_order
    )
    window_response = filtered_impulse_response(
        analytic_reference,
        analytic_reference,
        dt,
        filter_f0,
        filter_sigma,
        filter_order,
    )
    matched_window_response = minimize_response_difference(response, window_response)
    return np.real(response - matched_window_response)



from .profileAndIntensity import profile_analysis
from .trace_handler import MultiTraceHandler, TraceHandler
from .UVSpectrumAnalysis import (
    calibrate,
    load_calibration_data,
    plot_spectra,
    plot_spectra_UVsp,
    read_csd_file,
    read_spectrometer_excel,
    smooth,
    tukey_f,
    tukey_window,
)
from .VISSpectrumAnalysis import (
    MultiSpectrumHandler,
    SpectrumHandler,
    asymmetric_tukey_f,
    asymmetric_tukey_window,
    eliminate_outliers,
    read_spectrum_maya,
    read_spectrum_ocean_optics,
)

__all__ = [
    "profile_analysis",
    "TraceHandler",
    "MultiTraceHandler",
    "load_calibration_data",
    "smooth",
    "tukey_f",
    "tukey_window",
    "read_csd_file",
    "plot_spectra",
    "read_spectrometer_excel",
    "calibrate",
    "plot_spectra_UVsp",
    "eliminate_outliers",
    "read_spectrum_maya",
    "read_spectrum_ocean_optics",
    "asymmetric_tukey_f",
    "asymmetric_tukey_window",
    "SpectrumHandler",
    "MultiSpectrumHandler",
]
